#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.


import logging
import os
import sys
import tempfile
from subprocess import PIPE, CalledProcessError, check_call  # nosec
from typing import List, Optional

from onefuzztypes.models import NotificationConfig
from onefuzztypes.primitives import PoolName

from onefuzz.api import Command, Onefuzz
from onefuzz.cli import execute_api

SANITIZERS = ["address", "dataflow", "memory", "undefined"]


class Ossfuzz(Command):
    def build(self, project: str, sanitizer: str) -> None:
        """Build the latest oss-fuzz target"""
        self.logger.info("building %s:%s", project, sanitizer)
        cmd = [
            "docker",
            "run",
            "--rm",
            "-ti",
            "-e",
            "SANITIZER=%s" % sanitizer,
            "--mount",
            "src=%s,target=/out,type=bind" % os.getcwd(),
            "gcr.io/oss-fuzz/%s" % project,
            "compile",
        ]

        check_call(cmd, stderr=PIPE, stdout=PIPE)

    def fuzz(
        self,
        project: str,
        build: str,
        pool: PoolName,
        sanitizers: Optional[List[str]] = None,
        notification_config: Optional[NotificationConfig] = None,
    ) -> None:
        """Build & Launch all of the libFuzzer targets for a given project"""

        if sanitizers is None:
            sanitizers = SANITIZERS

        for sanitizer in sanitizers:
            with tempfile.TemporaryDirectory() as tmpdir:
                os.chdir(tmpdir)
                try:
                    self.build(project, sanitizer)
                except CalledProcessError:
                    self.logger.warning("building %s:%s failed", project, sanitizer)
                    continue

                self.logger.info("launching %s:%s build:%s", project, sanitizer, build)
                self.onefuzz.template.ossfuzz.libfuzzer(
                    project,
                    "%s:%s" % (sanitizer, build),
                    pool,
                    max_target_count=0,
                    sync_inputs=True,
                    notification_config=notification_config,
                )

    def stop(self, project: str) -> None:
        for job in self.onefuzz.jobs.list():
            if job.config.project != project:
                continue
            if job.config.build != "base":
                continue
            self.logger.info("stopping %s: %s", job.job_id, job.state)
            self.onefuzz.jobs.delete(job.job_id)


def main() -> int:
    return execute_api(
        Ossfuzz(Onefuzz(), logging.getLogger("ossfuzz")), [Command], "0.0.1"
    )


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
from typing import Any, Dict, List

from onefuzztypes.enums import ContainerType
from onefuzztypes.job_templates import JobTemplateConfig, JobTemplateRequest
from onefuzztypes.models import Job, TaskContainers

from ..api import Endpoint
from .job_monitor import JobMonitor


def container_type_name(container_type: ContainerType) -> str:
    return container_type.name + "_dir"


class TemplateSubmitHandler(Endpoint):
    """Submit Job Template"""

    _endpoint = "job_templates"

    def _process_containers(
        self, request: JobTemplateRequest, args: Dict[str, Any]
    ) -> None:
        """Create containers based on the argparse args"""

        for container in request.containers:
            directory_arg = container_type_name(container.type)
            self.onefuzz.logger.info("creating container: %s", container.name)
            self.onefuzz.containers.create(
                container.name, metadata={"container_type": container.type.name}
            )

            if directory_arg in args and args[directory_arg] is not None:
                self.onefuzz.logger.info(
                    "uploading %s to %s", args[directory_arg], container.name
                )
                self.onefuzz.containers.files.upload_dir(
                    container.name, args[directory_arg]
                )
            elif container.type == ContainerType.setup and "target_exe" in args:
                # This is isn't "declarative", but models our existing paths for
                # templates.

                target_exe = args["target_exe"]
                if target_exe is None:
                    continue
                self.onefuzz.logger.info(
                    "uploading %s to %s", target_exe, container.name
                )
                self.onefuzz.containers.files.upload_file(container.name, target_exe)

                pdb_path = os.path.splitext(target_exe)[0] + ".pdb"
                if os.path.exists(pdb_path):
                    self.onefuzz.containers.files.upload_file(container.name, pdb_path)

    def _define_missing_containers(
        self, config: JobTemplateConfig, request: JobTemplateRequest
    ) -> None:

        for container_type in config.containers:
            seen = False
            for container in request.containers:
                if container_type == container.type:
                    seen = True
            if not seen:
                if not isinstance(request.user_fields["project"], str):
                    raise TypeError
                if not isinstance(request.user_fields["name"], str):
                    raise TypeError
                if not isinstance(request.user_fields["build"], str):
                    raise TypeError
                container_name = self.onefuzz.utils.build_container_name(
                    container_type=container_type,
                    project=request.user_fields["project"],
                    name=request.user_fields["name"],
                    build=request.user_fields["build"],
                    platform=config.os,
                )
                request.containers.append(
                    TaskContainers(name=container_name, type=container_type)
                )

    def _submit(self, request: JobTemplateRequest) -> Job:
        self.onefuzz.logger.debug("submitting request: %s", request)
        return self._req_model(
            "POST", Job, data=request, alternate_endpoint=self._endpoint
        )

    def _execute_request(
        self,
        config: JobTemplateConfig,
        request: JobTemplateRequest,
        args: Dict[str, Any],
    ) -> Job:
        self._define_missing_containers(config, request)
        self._process_containers(request, args)
        return self._submit(request)

    def _convert_container_args(
        self, config: JobTemplateConfig, args: Dict[str, Any]
    ) -> List[TaskContainers]:
        """Convert the job template into a list of containers"""

        containers = []
        container_names = args["container_names"]
        if container_names is None:
            container_names = {}

        for container_type in config.containers:
            if container_type in container_names:
                container_name = container_names[container_type]
                containers.append(
                    TaskContainers(name=container_name, type=container_type)
                )
        return containers

    def _convert_args(
        self, config: JobTemplateConfig, args: Dict[str, Any]
    ) -> JobTemplateRequest:
        """convert arguments from argparse into a JobTemplateRequest"""

        user_fields = {}
        for field in config.user_fields:
            value = None
            if field.name in args:
                value = args[field.name]
            elif field.name in args["parameters"]:
                value = args["parameters"][field.name]
            elif field.required:
                raise Exception("missing field: %s" % field.name)

            if field.name == "target_exe" and isinstance(value, str):
                value = os.path.basename(value)

            if value is not None:
                user_fields[field.name] = value

        containers = self._convert_container_args(config, args)

        request = JobTemplateRequest(
            name=config.name, user_fields=user_fields, containers=containers
        )
        return request

    def _execute(
        self,
        config: JobTemplateConfig,
        args: Dict[str, Any],
        *,
        wait_for_running: bool,
    ) -> Job:
        """Convert argparse args into a JobTemplateRequest and submit it"""
        self.onefuzz.logger.debug("building: %s", config.name)
        request = self._convert_args(config, args)
        job = self._execute_request(config, request, args)

        JobMonitor(self.onefuzz, job).wait(wait_for_running=wait_for_running)
        return job



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json
import logging
import os
import pkgutil
import re
import subprocess  # nosec
import uuid
from enum import Enum
from shutil import which
from typing import Callable, Dict, List, Optional, Tuple, Type, TypeVar
from uuid import UUID

import semver
from memoization import cached
from onefuzztypes import (
    enums,
    events,
    models,
    primitives,
    requests,
    responses,
    webhooks,
)
from pydantic import BaseModel
from six.moves import input  # workaround for static analysis

from .__version__ import __version__
from .azcopy import azcopy_sync
from .backend import Backend, BackendConfig, ContainerWrapper, wait
from .ssh import build_ssh_command, ssh_connect, temp_file

UUID_EXPANSION = TypeVar("UUID_EXPANSION", UUID, str)

DEFAULT = BackendConfig(
    authority="https://login.microsoftonline.com/72f988bf-86f1-41af-91ab-2d7cd011db47",
    client_id="72f1562a-8c0c-41ea-beb9-fa2b71c80134",
)

# This was generated randomly and should be preserved moving forwards
ONEFUZZ_GUID_NAMESPACE = uuid.UUID("27f25e3f-6544-4b69-b309-9b096c5a9cbc")

ONE_HOUR_IN_SECONDS = 3600

DEFAULT_LINUX_IMAGE = "Canonical:UbuntuServer:18.04-LTS:latest"
DEFAULT_WINDOWS_IMAGE = "MicrosoftWindowsDesktop:Windows-10:20h2-pro:latest"

REPRO_SSH_FORWARD = "1337:127.0.0.1:1337"

UUID_RE = r"^[a-f0-9]{8}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{12}\Z"


class PreviewFeature(Enum):
    job_templates = "job_templates"


def is_uuid(value: str) -> bool:
    return bool(re.match(UUID_RE, value))


A = TypeVar("A", bound=BaseModel)


def _wsl_path(path: str) -> str:
    if which("wslpath"):
        # security note: path should always be a temporary path constructed by
        # this library
        return (
            subprocess.check_output(["wslpath", "-w", path]).decode().strip()  # nosec
        )
    return path


def user_confirmation(message: str) -> bool:
    answer: Optional[str] = None
    while answer not in ["y", "n"]:
        answer = input(message).strip()

    if answer == "n":
        return False
    return True


class Endpoint:
    endpoint: str

    def __init__(self, onefuzz: "Onefuzz"):
        self.onefuzz = onefuzz
        self.logger = onefuzz.logger

    def _req_model(
        self,
        method: str,
        model: Type[A],
        *,
        data: Optional[BaseModel] = None,
        as_params: bool = False,
        alternate_endpoint: Optional[str] = None,
    ) -> A:
        endpoint = self.endpoint if alternate_endpoint is None else alternate_endpoint

        if as_params:
            response = self.onefuzz._backend.request(method, endpoint, params=data)
        else:
            response = self.onefuzz._backend.request(method, endpoint, json_data=data)

        return model.parse_obj(response)

    def _req_model_list(
        self,
        method: str,
        model: Type[A],
        *,
        data: Optional[BaseModel] = None,
        as_params: bool = False,
        alternate_endpoint: Optional[str] = None,
    ) -> List[A]:
        endpoint = self.endpoint if alternate_endpoint is None else alternate_endpoint

        if as_params:
            response = self.onefuzz._backend.request(method, endpoint, params=data)
        else:
            response = self.onefuzz._backend.request(method, endpoint, json_data=data)

        return [model.parse_obj(x) for x in response]

    def _disambiguate(
        self,
        name: str,
        value: str,
        check: Callable[[str], bool],
        func: Callable[[], List[str]],
    ) -> str:
        if check(value):
            return value

        self.logger.debug("expanding %s: %s", name, value)

        values = [x for x in func() if x.startswith(value)]
        if len(values) == 1:
            return values[0]

        if len(values) > 1:
            if value in values:
                return value
            raise Exception(
                "%s expands to multiple values - %s: %s"
                % (name, value, ",".join(values))
            )

        raise Exception("Unable to find %s based on prefix: %s" % (name, value))

    def _disambiguate_uuid(
        self,
        name: str,
        value: UUID_EXPANSION,
        func: Callable[[], List[str]],
    ) -> UUID:
        if isinstance(value, UUID):
            return value
        return UUID(self._disambiguate(name, value, is_uuid, func))


class Files(Endpoint):
    """Interact with files within a container"""

    endpoint = "files"

    @cached(ttl=ONE_HOUR_IN_SECONDS)
    def _get_client(self, container: primitives.Container) -> ContainerWrapper:
        sas = self.onefuzz.containers.get(container).sas_url
        return ContainerWrapper(sas)

    def list(
        self, container: primitives.Container, prefix: Optional[str] = None
    ) -> models.Files:
        """Get a list of files in a container"""
        self.logger.debug("listing files in container: %s", container)
        client = self._get_client(container)
        return models.Files(files=client.list_blobs(name_starts_with=prefix))

    def delete(self, container: primitives.Container, filename: str) -> None:
        """delete a file from a container"""
        self.logger.debug("deleting in container: %s:%s", container, filename)
        client = self._get_client(container)
        client.delete_blob(filename)

    def get(self, container: primitives.Container, filename: str) -> bytes:
        """get a file from a container"""
        self.logger.debug("getting file from container: %s:%s", container, filename)
        client = self._get_client(container)
        downloaded = client.download_blob(filename)
        return downloaded

    def upload_file(
        self,
        container: primitives.Container,
        file_path: str,
        blob_name: Optional[str] = None,
    ) -> None:
        """uploads a file to a container"""
        if not blob_name:
            # Default blob name to file basename. This means that the file data will be
            # written to the "root" of the container, if simulating a directory tree.
            blob_name = os.path.basename(file_path)

        self.logger.debug(
            "uploading file to container %s:%s (blob_name: %s)",
            container,
            file_path,
            blob_name,
        )

        client = self._get_client(container)
        client.upload_file(file_path, blob_name)

    def upload_dir(
        self, container: primitives.Container, dir_path: primitives.Directory
    ) -> None:
        """uploads a directory to a container"""

        self.logger.debug("uploading directory to container %s:%s", container, dir_path)

        client = self._get_client(container)
        client.upload_dir(dir_path)

    def download_dir(
        self, container: primitives.Container, dir_path: primitives.Directory
    ) -> None:
        """downloads a container to a directory"""

        self.logger.debug(
            "downloading container to directory %s:%s", container, dir_path
        )

        client = self._get_client(container)
        client.download_dir(dir_path)


class Versions(Endpoint):
    """Onefuzz Instance"""

    def check(self, exact: bool = False) -> str:
        """Compare API and CLI versions for compatibility"""
        versions = self.onefuzz.info.get().versions
        api_str = versions["onefuzz"].version
        cli_str = __version__
        if exact:
            result = semver.compare(api_str, cli_str) == 0
        else:
            api = semver.VersionInfo.parse(api_str)
            cli = semver.VersionInfo.parse(cli_str)
            result = (
                api.major > 0 and api.major == cli.major and api.minor >= cli.minor
            ) or (
                api.major == 0
                and api.major == cli.major
                and api.minor == cli.minor
                and api.patch >= cli.patch
            )
            if cli_str == "0.0.0" and not result:
                self.logger.warning(
                    "ignoring compatibility check as the CLI was installed "
                    "from git.  api: %s cli: %s",
                    api_str,
                    cli_str,
                )
                result = True

        if not result:
            raise Exception(
                "incompatible versions.  api: %s cli: %s" % (api_str, cli_str)
            )

        return "compatible"


class Info(Endpoint):
    """Information about the OneFuzz instance"""

    endpoint = "info"

    def get(self) -> responses.Info:
        """Get information about the OneFuzz instance"""
        self.logger.debug("getting info")
        return self._req_model("GET", responses.Info)


class Webhooks(Endpoint):
    """Interact with Webhooks"""

    endpoint = "webhooks"

    def get(self, webhook_id: UUID_EXPANSION) -> webhooks.Webhook:
        """get a webhook"""

        webhook_id_expanded = self._disambiguate_uuid(
            "webhook_id", webhook_id, lambda: [str(x.webhook_id) for x in self.list()]
        )

        self.logger.debug("getting webhook: %s", webhook_id_expanded)
        return self._req_model(
            "GET",
            webhooks.Webhook,
            data=requests.WebhookSearch(webhook_id=webhook_id_expanded),
        )

    def list(self) -> List[webhooks.Webhook]:
        """list webhooks"""

        self.logger.debug("listing webhooks")
        return self._req_model_list(
            "GET",
            webhooks.Webhook,
            data=requests.WebhookSearch(),
        )

    def create(
        self,
        name: str,
        url: str,
        event_types: List[events.EventType],
        *,
        secret_token: Optional[str] = None,
    ) -> webhooks.Webhook:
        """Create a webhook"""
        self.logger.debug("creating webhook.  name: %s", name)
        return self._req_model(
            "POST",
            webhooks.Webhook,
            data=requests.WebhookCreate(
                name=name, url=url, event_types=event_types, secret_token=secret_token
            ),
        )

    def update(
        self,
        webhook_id: UUID_EXPANSION,
        *,
        name: Optional[str] = None,
        url: Optional[str] = None,
        event_types: Optional[List[events.EventType]] = None,
        secret_token: Optional[str] = None,
    ) -> webhooks.Webhook:
        """Update a webhook"""

        webhook_id_expanded = self._disambiguate_uuid(
            "webhook_id", webhook_id, lambda: [str(x.webhook_id) for x in self.list()]
        )

        self.logger.debug("updating webhook: %s", webhook_id_expanded)
        return self._req_model(
            "PATCH",
            webhooks.Webhook,
            data=requests.WebhookUpdate(
                webhook_id=webhook_id_expanded,
                name=name,
                url=url,
                event_types=event_types,
                secret_token=secret_token,
            ),
        )

    def delete(self, webhook_id: UUID_EXPANSION) -> responses.BoolResult:
        """Delete a webhook"""

        webhook_id_expanded = self._disambiguate_uuid(
            "webhook_id", webhook_id, lambda: [str(x.webhook_id) for x in self.list()]
        )

        return self._req_model(
            "DELETE",
            responses.BoolResult,
            data=requests.WebhookGet(webhook_id=webhook_id_expanded),
        )

    def ping(self, webhook_id: UUID_EXPANSION) -> events.EventPing:
        """ping a webhook"""

        webhook_id_expanded = self._disambiguate_uuid(
            "webhook_id", webhook_id, lambda: [str(x.webhook_id) for x in self.list()]
        )

        self.logger.debug("pinging webhook: %s", webhook_id_expanded)
        return self._req_model(
            "POST",
            events.EventPing,
            data=requests.WebhookGet(webhook_id=webhook_id_expanded),
            alternate_endpoint="webhooks/ping",
        )

    def logs(self, webhook_id: UUID_EXPANSION) -> List[webhooks.WebhookMessageLog]:
        """retreive webhook event log"""

        webhook_id_expanded = self._disambiguate_uuid(
            "webhook_id", webhook_id, lambda: [str(x.webhook_id) for x in self.list()]
        )

        self.logger.debug("pinging webhook: %s", webhook_id_expanded)
        return self._req_model_list(
            "POST",
            webhooks.WebhookMessageLog,
            data=requests.WebhookGet(webhook_id=webhook_id_expanded),
            alternate_endpoint="webhooks/logs",
        )


class Containers(Endpoint):
    """Interact with Onefuzz containers"""

    endpoint = "containers"

    def __init__(self, onefuzz: "Onefuzz"):
        super().__init__(onefuzz)
        self.files = Files(onefuzz)

    def get(self, name: str) -> responses.ContainerInfo:
        """Get a fully qualified SAS URL for a container"""
        self.logger.debug("get container: %s", name)
        return self._req_model(
            "GET", responses.ContainerInfo, data=requests.ContainerGet(name=name)
        )

    def create(
        self, name: str, metadata: Optional[Dict[str, str]] = None
    ) -> responses.ContainerInfo:
        """Create a storage container"""
        self.logger.debug("create container: %s", name)
        return self._req_model(
            "POST",
            responses.ContainerInfo,
            data=requests.ContainerCreate(name=name, metadata=metadata),
        )

    def delete(self, name: str) -> responses.BoolResult:
        """Delete a storage container"""
        self.logger.debug("delete container: %s", name)
        return self._req_model(
            "DELETE", responses.BoolResult, data=requests.ContainerDelete(name=name)
        )

    def list(self) -> List[responses.ContainerInfoBase]:
        """Get a list of containers"""
        self.logger.debug("list containers")
        return self._req_model_list("GET", responses.ContainerInfoBase)

    def reset(
        self,
        *,
        container_types: Optional[
            List[enums.ContainerType]
        ] = enums.ContainerType.reset_defaults(),
        yes: bool = False,
    ) -> None:
        """
        Reset containers by container type  (NOTE: This may cause unexpected issues with existing fuzzing jobs)
        """
        if not container_types:
            return

        message = "Confirm deleting container types: %s (specify y or n): " % (
            ",".join(x.name for x in container_types)
        )
        if not yes and not user_confirmation(message):
            self.logger.warning("not deleting containers")
            return

        for container in self.list():
            if (
                container.metadata
                and "container_type" in container.metadata
                and enums.ContainerType(container.metadata["container_type"])
                in container_types
            ):
                self.logger.info("removing container: %s", container.name)
                self.delete(container.name)


class Repro(Endpoint):
    """Interact with Reproduction VMs"""

    endpoint = "repro_vms"

    def get(self, vm_id: UUID_EXPANSION) -> models.Repro:
        """get information about a Reproduction VM"""
        vm_id_expanded = self._disambiguate_uuid(
            "vm_id", vm_id, lambda: [str(x.vm_id) for x in self.list()]
        )

        self.logger.debug("get repro vm: %s", vm_id_expanded)
        return self._req_model(
            "GET", models.Repro, data=requests.ReproGet(vm_id=vm_id_expanded)
        )

    def create(
        self, container: primitives.Container, path: str, duration: int = 24
    ) -> models.Repro:
        """Create a Reproduction VM from a Crash Report"""
        self.logger.info(
            "creating repro vm: %s %s (%d hours)", container, path, duration
        )
        return self._req_model(
            "POST",
            models.Repro,
            data=models.ReproConfig(container=container, path=path, duration=duration),
        )

    def delete(self, vm_id: UUID_EXPANSION) -> models.Repro:
        """Delete a Reproduction VM"""
        vm_id_expanded = self._disambiguate_uuid(
            "vm_id", vm_id, lambda: [str(x.vm_id) for x in self.list()]
        )

        self.logger.debug("deleting repro vm: %s", vm_id_expanded)
        return self._req_model(
            "DELETE", models.Repro, data=requests.ReproGet(vm_id=vm_id_expanded)
        )

    def list(self) -> List[models.Repro]:
        """List all VMs"""
        self.logger.debug("listing repro vms")
        return self._req_model_list("GET", models.Repro, data=requests.ReproGet())

    def _dbg_linux(
        self, repro: models.Repro, debug_command: Optional[str]
    ) -> Optional[str]:
        """Launch gdb with GDB script that includes 'target remote | ssh ...'"""

        if (
            repro.auth is None
            or repro.ip is None
            or repro.state != enums.VmState.running
        ):
            raise Exception("vm setup failed: %s" % repro.state)

        with build_ssh_command(
            repro.ip, repro.auth.private_key, command="-T"
        ) as ssh_cmd:

            gdb_script = [
                "target remote | %s sudo /onefuzz/bin/repro-stdout.sh"
                % " ".join(ssh_cmd)
            ]

            if debug_command:
                gdb_script += [debug_command, "quit"]

            with temp_file("gdb.script", "\n".join(gdb_script)) as gdb_script_path:
                dbg = ["gdb", "--silent", "--command", gdb_script_path]

                if debug_command:
                    dbg += ["--batch"]

                    try:
                        # security note: dbg is built from content coming from
                        # the server, which is trusted in this context.
                        return subprocess.run(  # nosec
                            dbg, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                        ).stdout.decode(errors="ignore")
                    except subprocess.CalledProcessError as err:
                        self.logger.error(
                            "debug failed: %s", err.output.decode(errors="ignore")
                        )
                        raise err
                else:
                    # security note: dbg is built from content coming from the
                    # server, which is trusted in this context.
                    subprocess.call(dbg)  # nosec
        return None

    def _dbg_windows(
        self, repro: models.Repro, debug_command: Optional[str]
    ) -> Optional[str]:
        """Setup an SSH tunnel, then connect via CDB over SSH tunnel"""

        if (
            repro.auth is None
            or repro.ip is None
            or repro.state != enums.VmState.running
        ):
            raise Exception("vm setup failed: %s" % repro.state)

        bind_all = which("wslpath") is not None and repro.os == enums.OS.windows
        proxy = "*:" + REPRO_SSH_FORWARD if bind_all else REPRO_SSH_FORWARD
        with ssh_connect(repro.ip, repro.auth.private_key, proxy=proxy):
            dbg = ["cdb.exe", "-remote", "tcp:port=1337,server=localhost"]
            if debug_command:
                dbg_script = [debug_command, "qq"]
                with temp_file("db.script", "\r\n".join(dbg_script)) as dbg_script_path:
                    dbg += ["-cf", _wsl_path(dbg_script_path)]

                    logging.debug("launching: %s", dbg)
                    try:
                        # security note: dbg is built from content coming from the server,
                        # which is trusted in this context.
                        return subprocess.run(  # nosec
                            dbg, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                        ).stdout.decode(errors="ignore")
                    except subprocess.CalledProcessError as err:
                        self.logger.error(
                            "debug failed: %s", err.output.decode(errors="ignore")
                        )
                        raise err
            else:
                logging.debug("launching: %s", dbg)
                # security note:  dbg is built from content coming from the
                # server, which is trusted in this context.
                subprocess.call(dbg)  # nosec

        return None

    def connect(
        self,
        vm_id: UUID_EXPANSION,
        delete_after_use: bool = False,
        debug_command: Optional[str] = None,
    ) -> Optional[str]:
        """Connect to an existing Reproduction VM"""

        self.logger.info("connecting to reproduction VM: %s", vm_id)

        if which("ssh") is None:
            raise Exception("unable to find ssh")

        def missing_os() -> Tuple[bool, str, models.Repro]:
            repro = self.get(vm_id)
            return (
                repro.os is not None,
                "waiting for os determination",
                repro,
            )

        repro = wait(missing_os)

        if repro.os == enums.OS.windows:
            if which("cdb.exe") is None:
                raise Exception("unable to find cdb.exe")
        if repro.os == enums.OS.linux:
            if which("gdb") is None:
                raise Exception("unable to find gdb")

        def func() -> Tuple[bool, str, models.Repro]:
            repro = self.get(vm_id)
            state = repro.state
            return (
                repro.auth is not None
                and repro.ip is not None
                and state not in [enums.VmState.init, enums.VmState.extensions_launch],
                "launching reproducing vm.  current state: %s" % state,
                repro,
            )

        repro = wait(func)

        result: Optional[str] = None

        if repro.os == enums.OS.windows:
            result = self._dbg_windows(repro, debug_command)
        elif repro.os == enums.OS.linux:
            result = self._dbg_linux(repro, debug_command)
        else:
            raise NotImplementedError

        if delete_after_use:
            self.logger.debug("deleting vm %s", repro.vm_id)
            self.delete(repro.vm_id)

        return result

    def create_and_connect(
        self,
        container: primitives.Container,
        path: str,
        duration: int = 24,
        delete_after_use: bool = False,
        debug_command: Optional[str] = None,
    ) -> Optional[str]:
        """Create and connect to a Reproduction VM"""
        repro = self.create(container, path, duration=duration)
        return self.connect(
            repro.vm_id, delete_after_use=delete_after_use, debug_command=debug_command
        )


class Notifications(Endpoint):
    """Interact with models.Notifications"""

    endpoint = "notifications"

    def create(
        self, container: primitives.Container, config: models.NotificationConfig
    ) -> models.Notification:
        """Create a notification based on a config file"""

        config = requests.NotificationCreate(container=container, config=config.config)
        return self._req_model("POST", models.Notification, data=config)

    def create_teams(
        self, container: primitives.Container, url: str
    ) -> models.Notification:
        """Create a Teams notification integration"""

        self.logger.debug("create teams notification integration: %s", container)

        config = models.NotificationConfig(config=models.TeamsTemplate(url=url))
        return self.create(container, config)

    def create_ado(
        self,
        container: primitives.Container,
        project: str,
        base_url: str,
        auth_token: str,
        work_item_type: str,
        unique_fields: List[str],
        comment: Optional[str] = None,
        fields: Optional[Dict[str, str]] = None,
        on_dup_increment: Optional[List[str]] = None,
        on_dup_comment: Optional[str] = None,
        on_dup_set_state: Optional[Dict[str, str]] = None,
        on_dup_fields: Optional[Dict[str, str]] = None,
    ) -> models.Notification:
        """Create an Azure DevOps notification integration"""

        self.logger.debug("create ado notification integration: %s", container)

        entry = models.NotificationConfig(
            config=models.ADOTemplate(
                base_url=base_url,
                auth_token=auth_token,
                project=project,
                type=work_item_type,
                comment=comment,
                unique_fields=unique_fields,
                ado_fields=fields,
                on_duplicate=models.ADODuplicateTemplate(
                    increment=on_dup_increment or [],
                    comment=on_dup_comment,
                    ado_fields=on_dup_fields or {},
                    set_state=on_dup_set_state or {},
                ),
            ),
        )
        return self.create(container, entry)

    def delete(self, notification_id: UUID_EXPANSION) -> models.Notification:
        """Delete a notification integration"""

        notification_id_expanded = self._disambiguate_uuid(
            "notification_id",
            notification_id,
            lambda: [str(x.notification_id) for x in self.list()],
        )

        self.logger.debug(
            "create notification integration: %s",
            notification_id_expanded,
        )
        return self._req_model(
            "DELETE",
            models.Notification,
            data=requests.NotificationGet(notification_id=notification_id_expanded),
        )

    def list(self) -> List[models.Notification]:
        """List notification integrations"""

        self.logger.debug("listing notification integrations")
        return self._req_model_list("GET", models.Notification)


class Tasks(Endpoint):
    """Interact with tasks"""

    endpoint = "tasks"

    def delete(self, task_id: UUID_EXPANSION) -> models.Task:
        """Stop an individual task"""

        task_id_expanded = self._disambiguate_uuid(
            "task_id", task_id, lambda: [str(x.task_id) for x in self.list()]
        )

        self.logger.debug("delete task: %s", task_id_expanded)

        return self._req_model(
            "DELETE", models.Task, data=requests.TaskGet(task_id=task_id_expanded)
        )

    def get(self, task_id: UUID_EXPANSION) -> models.Task:
        """Get information about a task"""
        task_id_expanded = self._disambiguate_uuid(
            "task_id", task_id, lambda: [str(x.task_id) for x in self.list()]
        )

        self.logger.debug("get task: %s", task_id_expanded)

        return self._req_model(
            "GET", models.Task, data=requests.TaskGet(task_id=task_id_expanded)
        )

    def create_with_config(self, config: models.TaskConfig) -> models.Task:
        """Create a Task using TaskConfig"""

        return self._req_model("POST", models.Task, data=config)

    def create(
        self,
        job_id: UUID_EXPANSION,
        task_type: enums.TaskType,
        target_exe: str,
        containers: List[Tuple[enums.ContainerType, primitives.Container]],
        *,
        analyzer_env: Optional[Dict[str, str]] = None,
        analyzer_exe: Optional[str] = None,
        analyzer_options: Optional[List[str]] = None,
        check_asan_log: bool = False,
        check_debugger: bool = True,
        check_retry_count: Optional[int] = None,
        check_fuzzer_help: Optional[bool] = None,
        expect_crash_on_failure: Optional[bool] = None,
        debug: Optional[List[enums.TaskDebugFlag]] = None,
        duration: int = 24,
        ensemble_sync_delay: Optional[int] = None,
        generator_exe: Optional[str] = None,
        generator_options: Optional[List[str]] = None,
        pool_name: primitives.PoolName,
        prereq_tasks: Optional[List[UUID]] = None,
        reboot_after_setup: bool = False,
        rename_output: bool = False,
        stats_file: Optional[str] = None,
        stats_format: Optional[enums.StatsFormat] = None,
        supervisor_env: Optional[Dict[str, str]] = None,
        supervisor_exe: Optional[str] = None,
        supervisor_input_marker: Optional[str] = None,
        supervisor_options: Optional[List[str]] = None,
        tags: Optional[Dict[str, str]] = None,
        task_wait_for_files: Optional[enums.ContainerType] = None,
        target_env: Optional[Dict[str, str]] = None,
        target_options: Optional[List[str]] = None,
        target_options_merge: bool = False,
        target_timeout: Optional[int] = None,
        target_workers: Optional[int] = None,
        vm_count: int = 1,
        preserve_existing_outputs: bool = False,
        colocate: bool = False,
        report_list: Optional[List[str]] = None,
        minimized_stack_depth: Optional[int] = None,
        coverage_filter: Optional[str] = None,
    ) -> models.Task:
        """
        Create a task

        :param bool ensemble_sync_delay: Specify duration between
            syncing inputs during ensemble fuzzing (0 to disable).
        """

        self.logger.debug("creating task: %s", task_type)

        job_id_expanded = self._disambiguate_uuid(
            "job_id",
            job_id,
            lambda: [str(x.job_id) for x in self.onefuzz.jobs.list()],
        )

        if tags is None:
            tags = {}

        containers_submit = []
        for (container_type, container) in containers:
            containers_submit.append(
                models.TaskContainers(name=container, type=container_type)
            )

        config = models.TaskConfig(
            containers=containers_submit,
            debug=debug,
            job_id=job_id_expanded,
            pool=models.TaskPool(count=vm_count, pool_name=pool_name),
            prereq_tasks=prereq_tasks,
            tags=tags,
            colocate=colocate,
            task=models.TaskDetails(
                analyzer_env=analyzer_env,
                analyzer_exe=analyzer_exe,
                analyzer_options=analyzer_options,
                check_asan_log=check_asan_log,
                check_debugger=check_debugger,
                check_retry_count=check_retry_count,
                check_fuzzer_help=check_fuzzer_help,
                expect_crash_on_failure=expect_crash_on_failure,
                duration=duration,
                ensemble_sync_delay=ensemble_sync_delay,
                generator_exe=generator_exe,
                generator_options=generator_options,
                reboot_after_setup=reboot_after_setup,
                rename_output=rename_output,
                stats_file=stats_file,
                stats_format=stats_format,
                supervisor_env=supervisor_env,
                supervisor_exe=supervisor_exe,
                supervisor_input_marker=supervisor_input_marker,
                supervisor_options=supervisor_options,
                target_env=target_env,
                target_exe=target_exe,
                target_options=target_options,
                target_options_merge=target_options_merge,
                target_timeout=target_timeout,
                target_workers=target_workers,
                type=task_type,
                wait_for_files=task_wait_for_files,
                report_list=report_list,
                preserve_existing_outputs=preserve_existing_outputs,
                minimized_stack_depth=minimized_stack_depth,
                coverage_filter=coverage_filter,
            ),
        )

        return self.create_with_config(config)

    def list(
        self,
        job_id: Optional[UUID_EXPANSION] = None,
        state: Optional[List[enums.TaskState]] = enums.TaskState.available(),
    ) -> List[models.Task]:
        """Get information about all tasks"""
        self.logger.debug("list tasks")
        job_id_expanded: Optional[UUID] = None

        if job_id is not None:
            job_id_expanded = self._disambiguate_uuid(
                "job_id",
                job_id,
                lambda: [str(x.job_id) for x in self.onefuzz.jobs.list()],
            )

        return self._req_model_list(
            "GET",
            models.Task,
            data=requests.TaskSearch(job_id=job_id_expanded, state=state),
        )


class JobContainers(Endpoint):
    """Interact with Containers used within tasks in a Job"""

    endpoint = "jobs"

    def list(
        self,
        job_id: UUID_EXPANSION,
        container_type: Optional[
            enums.ContainerType
        ] = enums.ContainerType.unique_reports,
    ) -> Dict[str, List[str]]:
        """
        List the files for all of the containers of a given container type
        for the specified job
        """
        containers = set()
        tasks = self.onefuzz.tasks.list(job_id=job_id, state=[])
        for task in tasks:
            containers.update(
                set(x.name for x in task.config.containers if x.type == container_type)
            )

        results: Dict[str, List[str]] = {}
        for container in containers:
            results[container] = self.onefuzz.containers.files.list(container).files
        return results

    def download(
        self, job_id: UUID_EXPANSION, *, output: Optional[primitives.Directory] = None
    ) -> None:
        to_download = {}
        tasks = self.onefuzz.tasks.list(job_id=job_id, state=None)
        if not tasks:
            raise Exception("no tasks with job_id:%s" % job_id)

        for task in tasks:
            for container in task.config.containers:
                info = self.onefuzz.containers.get(container.name)
                name = os.path.join(container.type.name, container.name)
                to_download[name] = info.sas_url

        if output is None:
            output = primitives.Directory(os.getcwd())

        for name in to_download:
            outdir = os.path.join(output, name)
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            self.logger.info("downloading: %s", name)
            # security note: the src for azcopy comes from the server which is
            # trusted in this context, while the destination is provided by the
            # user
            azcopy_sync(to_download[name], outdir)

    def delete(
        self,
        job_id: UUID_EXPANSION,
        *,
        only_job_specific: bool = True,
        dryrun: bool = False,
    ) -> None:
        SAFE_TO_REMOVE = [
            enums.ContainerType.crashes,
            enums.ContainerType.setup,
            enums.ContainerType.inputs,
            enums.ContainerType.reports,
            enums.ContainerType.unique_inputs,
            enums.ContainerType.unique_reports,
            enums.ContainerType.no_repro,
            enums.ContainerType.analysis,
            enums.ContainerType.coverage,
            enums.ContainerType.readonly_inputs,
            enums.ContainerType.regression_reports,
        ]

        job = self.onefuzz.jobs.get(job_id)
        containers = set()
        to_delete = set()
        for task in self.onefuzz.jobs.tasks.list(job_id=job.job_id):
            for container in task.config.containers:
                containers.add(container.name)
                if container.type not in SAFE_TO_REMOVE:
                    continue
                elif not only_job_specific:
                    to_delete.add(container.name)
                elif only_job_specific and (
                    self.onefuzz.utils.build_container_name(
                        container_type=container.type,
                        project=job.config.project,
                        name=job.config.name,
                        build=job.config.build,
                        platform=task.os,
                    )
                    == container.name
                ):
                    to_delete.add(container.name)

        to_keep = containers - to_delete
        for container_name in to_keep:
            self.logger.info("not removing: %s", container_name)

        for container_name in to_delete:
            if dryrun:
                self.logger.info("container would be deleted: %s", container_name)
            elif self.onefuzz.containers.delete(container_name).result:
                self.logger.info("removed container: %s", container_name)
            else:
                self.logger.info("container already removed: %s", container_name)


class JobTasks(Endpoint):
    """Interact with tasks within a job"""

    endpoint = "jobs"

    def list(self, job_id: UUID_EXPANSION) -> List[models.Task]:
        """List all of the tasks for a given job"""
        return self.onefuzz.tasks.list(job_id=job_id, state=[])


class Jobs(Endpoint):
    """Interact with Jobs"""

    endpoint = "jobs"

    def __init__(self, onefuzz: "Onefuzz"):
        super().__init__(onefuzz)
        self.containers = JobContainers(onefuzz)
        self.tasks = JobTasks(onefuzz)

    def delete(self, job_id: UUID_EXPANSION) -> models.Job:

        """Stop a job and all tasks that make up a job"""
        job_id_expanded = self._disambiguate_uuid(
            "job_id", job_id, lambda: [str(x.job_id) for x in self.list()]
        )

        self.logger.debug("delete job: %s", job_id_expanded)
        return self._req_model(
            "DELETE", models.Job, data=requests.JobGet(job_id=job_id_expanded)
        )

    def get(self, job_id: UUID_EXPANSION) -> models.Job:
        """Get information about a specific job"""
        job_id_expanded = self._disambiguate_uuid(
            "job_id", job_id, lambda: [str(x.job_id) for x in self.list()]
        )
        self.logger.debug("get job: %s", job_id_expanded)
        job = self._req_model(
            "GET", models.Job, data=requests.JobGet(job_id=job_id_expanded)
        )
        return job

    def create_with_config(self, config: models.JobConfig) -> models.Job:
        """Create a job"""
        self.logger.debug(
            "create job: project:%s name:%s build:%s",
            config.project,
            config.name,
            config.build,
        )
        return self._req_model(
            "POST",
            models.Job,
            data=config,
        )

    def create(
        self, project: str, name: str, build: str, duration: int = 24
    ) -> models.Job:
        """Create a job"""
        return self.create_with_config(
            models.JobConfig(project=project, name=name, build=build, duration=duration)
        )

    def list(
        self,
        job_state: Optional[List[enums.JobState]] = enums.JobState.available(),
    ) -> List[models.Job]:
        """Get information about all jobs"""
        self.logger.debug("list jobs")
        return self._req_model_list(
            "GET", models.Job, data=requests.JobSearch(state=job_state)
        )


class Pool(Endpoint):
    """Interact with worker pools"""

    endpoint = "pool"

    def create(
        self,
        name: str,
        os: enums.OS,
        client_id: Optional[UUID] = None,
        *,
        unmanaged: bool = False,
        arch: enums.Architecture = enums.Architecture.x86_64,
    ) -> models.Pool:
        """
        Create a worker pool

        :param str name: Name of the worker-pool
        """
        self.logger.debug("create worker pool")
        managed = not unmanaged

        return self._req_model(
            "POST",
            models.Pool,
            data=requests.PoolCreate(
                name=name, os=os, arch=arch, managed=managed, client_id=client_id
            ),
        )

    def get_config(self, pool_name: primitives.PoolName) -> models.AgentConfig:
        """Get the agent configuration for the pool"""

        pool = self.get(pool_name)

        if pool.config is None:
            raise Exception("Missing AgentConfig in response")

        config = pool.config
        config.client_credentials = models.ClientCredentials(  # nosec - bandit consider this a hard coded password
            client_id=pool.client_id,
            client_secret="<client secret>",
        )

        return config

    def shutdown(self, name: str, *, now: bool = False) -> responses.BoolResult:
        expanded_name = self._disambiguate(
            "name", name, lambda x: False, lambda: [x.name for x in self.list()]
        )

        self.logger.debug("shutdown worker pool: %s (now: %s)", expanded_name, now)
        return self._req_model(
            "DELETE",
            responses.BoolResult,
            data=requests.PoolStop(name=expanded_name, now=now),
        )

    def get(self, name: str) -> models.Pool:
        self.logger.debug("get details on a specific pool")
        expanded_name = self._disambiguate(
            "pool name", name, lambda x: False, lambda: [x.name for x in self.list()]
        )

        return self._req_model(
            "GET", models.Pool, data=requests.PoolSearch(name=expanded_name)
        )

    def list(
        self, *, state: Optional[List[enums.PoolState]] = None
    ) -> List[models.Pool]:
        self.logger.debug("list worker pools")
        return self._req_model_list(
            "GET", models.Pool, data=requests.PoolSearch(state=state)
        )


class Node(Endpoint):
    """Interact with nodes"""

    endpoint = "node"

    def get(self, machine_id: UUID_EXPANSION) -> models.Node:
        self.logger.debug("get node: %s", machine_id)
        machine_id_expanded = self._disambiguate_uuid(
            "machine_id",
            machine_id,
            lambda: [str(x.machine_id) for x in self.list()],
        )

        return self._req_model(
            "GET", models.Node, data=requests.NodeGet(machine_id=machine_id_expanded)
        )

    def halt(self, machine_id: UUID_EXPANSION) -> responses.BoolResult:
        self.logger.debug("halt node: %s", machine_id)
        machine_id_expanded = self._disambiguate_uuid(
            "machine_id",
            machine_id,
            lambda: [str(x.machine_id) for x in self.list()],
        )

        return self._req_model(
            "DELETE",
            responses.BoolResult,
            data=requests.NodeGet(machine_id=machine_id_expanded),
        )

    def reimage(self, machine_id: UUID_EXPANSION) -> responses.BoolResult:
        self.logger.debug("reimage node: %s", machine_id)
        machine_id_expanded = self._disambiguate_uuid(
            "machine_id",
            machine_id,
            lambda: [str(x.machine_id) for x in self.list()],
        )

        return self._req_model(
            "PATCH",
            responses.BoolResult,
            data=requests.NodeGet(machine_id=machine_id_expanded),
        )

    def update(
        self,
        machine_id: UUID_EXPANSION,
        *,
        debug_keep_node: Optional[bool] = None,
    ) -> responses.BoolResult:
        self.logger.debug("update node: %s", machine_id)
        machine_id_expanded = self._disambiguate_uuid(
            "machine_id",
            machine_id,
            lambda: [str(x.machine_id) for x in self.list()],
        )

        return self._req_model(
            "POST",
            responses.BoolResult,
            data=requests.NodeUpdate(
                machine_id=machine_id_expanded,
                debug_keep_node=debug_keep_node,
            ),
        )

    def list(
        self,
        *,
        state: Optional[List[enums.NodeState]] = None,
        scaleset_id: Optional[UUID_EXPANSION] = None,
        pool_name: Optional[primitives.PoolName] = None,
    ) -> List[models.Node]:
        self.logger.debug("list nodes")
        scaleset_id_expanded: Optional[UUID] = None

        if pool_name is not None:
            pool_name = primitives.PoolName(
                self._disambiguate(
                    "name",
                    str(pool_name),
                    lambda x: False,
                    lambda: [x.name for x in self.onefuzz.pools.list()],
                )
            )

        if scaleset_id is not None:
            scaleset_id_expanded = self._disambiguate_uuid(
                "scaleset_id",
                scaleset_id,
                lambda: [str(x.scaleset_id) for x in self.onefuzz.scalesets.list()],
            )

        return self._req_model_list(
            "GET",
            models.Node,
            data=requests.NodeSearch(
                scaleset_id=scaleset_id_expanded, state=state, pool_name=pool_name
            ),
        )

    def add_ssh_key(
        self, machine_id: UUID_EXPANSION, *, public_key: str
    ) -> responses.BoolResult:
        self.logger.debug("add ssh public key to node: %s", machine_id)
        machine_id_expanded = self._disambiguate_uuid(
            "machine_id",
            machine_id,
            lambda: [str(x.machine_id) for x in self.list()],
        )

        return self._req_model(
            "POST",
            responses.BoolResult,
            data=requests.NodeAddSshKey(
                machine_id=machine_id_expanded,
                public_key=public_key,
            ),
            alternate_endpoint="node/add_ssh_key",
        )


class Scaleset(Endpoint):
    """Interact with managed scaleset pools"""

    endpoint = "scaleset"

    def _expand_scaleset_machine(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        *,
        include_auth: bool = False,
    ) -> Tuple[models.Scaleset, UUID]:
        scaleset = self.get(scaleset_id, include_auth=include_auth)

        if scaleset.nodes is None:
            raise Exception("no nodes defined in scaleset")
        nodes = scaleset.nodes

        machine_id_expanded = self._disambiguate_uuid(
            "machine_id", machine_id, lambda: [str(x.machine_id) for x in nodes]
        )
        return (scaleset, machine_id_expanded)

    def create(
        self,
        pool_name: primitives.PoolName,
        size: int,
        *,
        image: Optional[str] = None,
        vm_sku: Optional[str] = "Standard_D2s_v3",
        region: Optional[primitives.Region] = None,
        spot_instances: bool = False,
        ephemeral_os_disks: bool = False,
        tags: Optional[Dict[str, str]] = None,
    ) -> models.Scaleset:
        self.logger.debug("create scaleset")

        if tags is None:
            tags = {}

        if image is None:
            pool = self.onefuzz.pools.get(pool_name)
            if pool.os == enums.OS.linux:
                image = DEFAULT_LINUX_IMAGE
            elif pool.os == enums.OS.windows:
                image = DEFAULT_WINDOWS_IMAGE
            else:
                raise NotImplementedError

        return self._req_model(
            "POST",
            models.Scaleset,
            data=requests.ScalesetCreate(
                pool_name=pool_name,
                vm_sku=vm_sku,
                image=image,
                region=region,
                size=size,
                spot_instances=spot_instances,
                ephemeral_os_disks=ephemeral_os_disks,
                tags=tags,
            ),
        )

    def shutdown(
        self, scaleset_id: UUID_EXPANSION, *, now: bool = False
    ) -> responses.BoolResult:
        scaleset_id_expanded = self._disambiguate_uuid(
            "scaleset_id",
            scaleset_id,
            lambda: [str(x.scaleset_id) for x in self.list()],
        )

        self.logger.debug("shutdown scaleset: %s (now: %s)", scaleset_id_expanded, now)
        return self._req_model(
            "DELETE",
            responses.BoolResult,
            data=requests.ScalesetStop(scaleset_id=scaleset_id_expanded, now=now),
        )

    def get(
        self, scaleset_id: UUID_EXPANSION, *, include_auth: bool = False
    ) -> models.Scaleset:
        self.logger.debug("get scaleset: %s", scaleset_id)
        scaleset_id_expanded = self._disambiguate_uuid(
            "scaleset_id",
            scaleset_id,
            lambda: [str(x.scaleset_id) for x in self.list()],
        )

        return self._req_model(
            "GET",
            models.Scaleset,
            data=requests.ScalesetSearch(
                scaleset_id=scaleset_id_expanded, include_auth=include_auth
            ),
        )

    def update(
        self, scaleset_id: UUID_EXPANSION, *, size: Optional[int] = None
    ) -> models.Scaleset:
        self.logger.debug("update scaleset: %s", scaleset_id)
        scaleset_id_expanded = self._disambiguate_uuid(
            "scaleset_id",
            scaleset_id,
            lambda: [str(x.scaleset_id) for x in self.list()],
        )

        return self._req_model(
            "PATCH",
            models.Scaleset,
            data=requests.ScalesetUpdate(scaleset_id=scaleset_id_expanded, size=size),
        )

    def list(
        self,
        *,
        state: Optional[List[enums.ScalesetState]] = None,
    ) -> List[models.Scaleset]:
        self.logger.debug("list scalesets")
        return self._req_model_list(
            "GET", models.Scaleset, data=requests.ScalesetSearch(state=state)
        )


class ScalesetProxy(Endpoint):
    """Interact with Scaleset Proxies (NOTE: This API is unstable)"""

    endpoint = "proxy"

    def delete(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        *,
        dst_port: Optional[int] = None,
    ) -> responses.BoolResult:
        """Stop a proxy node"""

        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(scaleset_id, machine_id)

        self.logger.debug(
            "delete proxy: %s:%d %d",
            scaleset.scaleset_id,
            machine_id_expanded,
            dst_port,
        )
        return self._req_model(
            "DELETE",
            responses.BoolResult,
            data=requests.ProxyDelete(
                scaleset_id=scaleset.scaleset_id,
                machine_id=machine_id_expanded,
                dst_port=dst_port,
            ),
        )

    def reset(self, region: primitives.Region) -> responses.BoolResult:
        """Reset the proxy for an existing region"""

        return self._req_model(
            "PATCH", responses.BoolResult, data=requests.ProxyReset(region=region)
        )

    def get(
        self, scaleset_id: UUID_EXPANSION, machine_id: UUID_EXPANSION, dst_port: int
    ) -> responses.ProxyGetResult:
        """Get information about a specific job"""
        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(scaleset_id, machine_id)

        self.logger.debug(
            "get proxy: %s:%d:%d", scaleset.scaleset_id, machine_id_expanded, dst_port
        )
        proxy = self._req_model(
            "GET",
            responses.ProxyGetResult,
            data=requests.ProxyGet(
                scaleset_id=scaleset.scaleset_id,
                machine_id=machine_id_expanded,
                dst_port=dst_port,
            ),
        )
        return proxy

    def create(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        dst_port: int,
        *,
        duration: Optional[int] = 1,
    ) -> responses.ProxyGetResult:
        """Create a proxy"""
        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(scaleset_id, machine_id)

        self.logger.debug(
            "create proxy: %s:%s %d",
            scaleset.scaleset_id,
            machine_id_expanded,
            dst_port,
        )
        return self._req_model(
            "POST",
            responses.ProxyGetResult,
            data=requests.ProxyCreate(
                scaleset_id=scaleset.scaleset_id,
                machine_id=machine_id_expanded,
                dst_port=dst_port,
                duration=duration,
            ),
        )

    def list(self) -> responses.ProxyList:
        return self._req_model("GET", responses.ProxyList, data=requests.ProxyGet())


class Command:
    def __init__(self, onefuzz: "Onefuzz", logger: logging.Logger):
        self.onefuzz = onefuzz
        self.logger = logger


class Utils(Command):
    def namespaced_guid(
        self,
        project: str,
        name: Optional[str] = None,
        build: Optional[str] = None,
        platform: Optional[str] = None,
    ) -> uuid.UUID:
        identifiers = [project]
        if name is not None:
            identifiers.append(name)
        if build is not None:
            identifiers.append(build)
        if platform is not None:
            identifiers.append(platform)
        return uuid.uuid5(ONEFUZZ_GUID_NAMESPACE, ":".join(identifiers))

    def build_container_name(
        self,
        *,
        container_type: enums.ContainerType,
        project: str,
        name: str,
        build: str,
        platform: enums.OS,
    ) -> primitives.Container:
        if container_type in [enums.ContainerType.setup, enums.ContainerType.coverage]:
            guid = self.namespaced_guid(
                project,
                name,
                build=build,
                platform=platform.name,
            )
        elif container_type == enums.ContainerType.regression_reports:
            guid = self.namespaced_guid(
                project,
                name,
                build=build,
            )
        else:
            guid = self.namespaced_guid(project, name)

        return primitives.Container(
            "oft-%s-%s"
            % (
                container_type.name.replace("_", "-"),
                guid.hex,
            )
        )


class Onefuzz:
    def __init__(
        self, config_path: Optional[str] = None, token_path: Optional[str] = None
    ) -> None:
        self.logger = logging.getLogger("onefuzz")
        self._backend = Backend(
            config=DEFAULT, config_path=config_path, token_path=token_path
        )
        self.containers = Containers(self)
        self.repro = Repro(self)
        self.notifications = Notifications(self)
        self.tasks = Tasks(self)
        self.jobs = Jobs(self)
        self.versions = Versions(self)
        self.info = Info(self)
        self.scaleset_proxy = ScalesetProxy(self)
        self.pools = Pool(self)
        self.scalesets = Scaleset(self)
        self.nodes = Node(self)
        self.webhooks = Webhooks(self)

        if self._backend.is_feature_enabled(PreviewFeature.job_templates.name):
            self.job_templates = JobTemplates(self)

        # these are externally developed cli modules
        self.template = Template(self, self.logger)
        self.debug = Debug(self, self.logger)
        self.status = Status(self, self.logger)
        self.utils = Utils(self, self.logger)

        self.__setup__()

    def __setup__(
        self,
        endpoint: Optional[str] = None,
        client_id: Optional[str] = None,
        authority: Optional[str] = None,
        tenant_domain: Optional[str] = None,
    ) -> None:

        if endpoint:
            self._backend.config.endpoint = endpoint
        if authority is not None:
            self._backend.config.authority = authority
        if client_id is not None:
            self._backend.config.client_id = client_id
        if tenant_domain is not None:
            self._backend.config.tenant_domain = tenant_domain

        if self._backend.is_feature_enabled(PreviewFeature.job_templates.name):
            self.job_templates._load_cache()

    def licenses(self) -> object:
        """Return third-party licenses used by this package"""
        data = pkgutil.get_data("onefuzz", "data/licenses.json")
        if data is None:
            raise Exception("missing licenses.json")
        return json.loads(data)

    def privacy_statement(self) -> bytes:
        """Return OneFuzz privacy statement"""
        data = pkgutil.get_data("onefuzz", "data/privacy.txt")
        if data is None:
            raise Exception("missing licenses.json")
        return data

    def logout(self) -> None:
        """Logout of Onefuzz"""
        self.logger.debug("logout")

        self._backend.logout()

    def login(self) -> str:
        """Login to Onefuzz"""

        # Rather than interacting MSAL directly, call a simple API which
        # actuates the login process
        self.info.get()

        # TODO: once job templates are out of preview, this should be enabled
        if self._backend.is_feature_enabled(PreviewFeature.job_templates.name):
            self.job_templates.refresh()
        return "succeeded"

    def config(
        self,
        endpoint: Optional[str] = None,
        authority: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        enable_feature: Optional[PreviewFeature] = None,
        tenant_domain: Optional[str] = None,
        reset: Optional[bool] = None,
    ) -> BackendConfig:
        """Configure onefuzz CLI"""
        self.logger.debug("set config")

        if reset:
            self._backend.config = BackendConfig(authority="", client_id="")

        if endpoint is not None:
            # The normal path for calling the API always uses the oauth2 workflow,
            # which the devicelogin can take upwards of 15 minutes to fail in
            # error cases.
            #
            # This check only happens on setting the configuration, as checking the
            # viability of the service on every call is prohibitively expensive.
            verify = self._backend.session.request("GET", endpoint)
            if verify.status_code != 401:
                self.logger.warning(
                    "This could be an invalid OneFuzz API endpoint: "
                    "Missing HTTP Authentication"
                )
            self._backend.config.endpoint = endpoint
        if authority is not None:
            self._backend.config.authority = authority
        if client_id is not None:
            self._backend.config.client_id = client_id
        if client_secret is not None:
            self._backend.config.client_secret = client_secret
        if enable_feature:
            self._backend.enable_feature(enable_feature.name)
        if tenant_domain is not None:
            self._backend.config.tenant_domain = tenant_domain
        self._backend.app = None
        self._backend.save_config()

        data = self._backend.config.copy(deep=True)
        if data.client_secret is not None:
            # replace existing secrets with "*** for user display
            data.client_secret = "***"  # nosec

        if not data.endpoint:
            self.logger.warning("endpoint not configured yet")

        return data

    def _delete_components(
        self,
        *,
        containers: bool = False,
        jobs: bool = False,
        notifications: bool = False,
        pools: bool = False,
        repros: bool = False,
        scalesets: bool = False,
        tasks: bool = False,
        webhooks: bool = False,
    ) -> None:
        if jobs:
            for job in self.jobs.list():
                self.logger.info("stopping job %s", job.job_id)
                self.jobs.delete(job.job_id)

        if tasks:
            for task in self.tasks.list():
                self.logger.info("stopping task %s", task.task_id)
                self.tasks.delete(task.task_id)

        if notifications:
            for notification in self.notifications.list():
                self.logger.info(
                    "stopping notification %s", notification.notification_id
                )
                self.notifications.delete(notification.notification_id)

        if repros:
            for vm in self.repro.list():
                self.repro.delete(str(vm.vm_id))

        if pools:
            for pool in self.pools.list():
                self.logger.info("stopping pool: %s", pool.name)
                self.pools.shutdown(pool.name, now=True)

        if scalesets:
            for scaleset in self.scalesets.list():
                self.logger.info("stopping scaleset: %s", scaleset.scaleset_id)
                self.scalesets.shutdown(scaleset.scaleset_id, now=True)

        if containers:
            self.containers.reset(yes=True)

        if webhooks:
            for webhook in self.webhooks.list():
                self.logger.info("removing webhook: %s", webhook.webhook_id)
                self.webhooks.delete(webhook.webhook_id)

    def reset(
        self,
        *,
        containers: bool = False,
        everything: bool = False,
        jobs: bool = False,
        notifications: bool = False,
        pools: bool = False,
        repros: bool = False,
        scalesets: bool = False,
        tasks: bool = False,
        webhooks: bool = False,
        yes: bool = False,
    ) -> None:
        """
        Resets onefuzz. Stops all jobs, notifications, and repro VMs.
        Specifying 'everything' will delete all containers, pools, and managed
        scalesets.

        :param bool containers: Delete all the containers.
        :param bool everything: Delete all containers, pools and managed scalesets.
        :param bool jobs: Stop all jobs.
        :param bool notifications: Stop all notifications.
        :param bool pools: Delete all pools.
        :param bool repros: Delete all repro vms.
        :param bool scalesets: Delete all managed scalesets.
        :param bool tasks: Stop all tasks.
        :param bool webhooks: Stop all webhooks.
        :param bool yes: Ignoring to specify "y" in prompt.
        """

        if everything:
            (
                containers,
                jobs,
                pools,
                notifications,
                repros,
                scalesets,
                tasks,
                webhooks,
            ) = (
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            )
        elif not any(
            [containers, jobs, pools, notifications, repros, scalesets, tasks]
        ):
            jobs, notifications, repros, tasks = True, True, True, True

        if containers and not (tasks or jobs):
            raise Exception("Resetting containers requires resetting jobs or tasks")

        to_delete = []
        argument_str = {
            "jobs",
            "tasks",
            "notifications",
            "pools",
            "scalesets",
            "repros",
            "containers",
            "webhooks",
        }
        for k, v in locals().items():
            if k in argument_str and v:
                to_delete.append(k)
        message = "Confirm stopping %s (specify y or n): " % (
            ", ".join(sorted(to_delete))
        )
        if not yes and not user_confirmation(message):
            self.logger.warning("not resetting")
            return

        self._delete_components(
            containers=containers,
            jobs=jobs,
            notifications=notifications,
            pools=pools,
            repros=repros,
            scalesets=scalesets,
            tasks=tasks,
            webhooks=webhooks,
        )

    def _warn_preview(self, feature: PreviewFeature) -> None:
        self.logger.warning(
            "%s are a preview-feature and may change in an upcoming release",
            feature.name,
        )


from .debug import Debug  # noqa: E402
from .job_templates.main import JobTemplates  # noqa: E402
from .status.cmd import Status  # noqa: E402
from .template import Template  # noqa: E402



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json
import logging
import os
import tempfile
import time
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from urllib.parse import urlparse
from uuid import UUID

import jmespath
from azure.applicationinsights import ApplicationInsightsDataClient
from azure.applicationinsights.models import QueryBody
from azure.common.client_factory import get_azure_cli_credentials
from onefuzztypes.enums import ContainerType, TaskType
from onefuzztypes.models import BlobRef, Job, NodeAssignment, Report, Task, TaskConfig
from onefuzztypes.primitives import Container, Directory, PoolName

from onefuzz.api import UUID_EXPANSION, Command, Onefuzz

from .backend import wait
from .rdp import rdp_connect
from .ssh import ssh_connect

EMPTY_SHA256 = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
ZERO_SHA256 = "0" * len(EMPTY_SHA256)
DAY_TIMESPAN = "PT24H"
HOUR_TIMESPAN = "PT1H"
DEFAULT_TAIL_DELAY = 10.0


class DebugRepro(Command):
    """Debug repro instances"""

    def _disambiguate(self, vm_id: UUID_EXPANSION) -> str:
        return str(
            self.onefuzz.repro._disambiguate_uuid(
                "vm_id",
                vm_id,
                lambda: [str(x.vm_id) for x in self.onefuzz.repro.list()],
            )
        )

    def _info(self) -> Tuple[str, str]:
        info = self.onefuzz.info.get()
        return info.resource_group, info.subscription

    def ssh(self, vm_id: str) -> None:
        vm_id = self._disambiguate(vm_id)
        repro = self.onefuzz.repro.get(vm_id)
        if repro.ip is None:
            raise Exception("missing IP: %s" % repro)
        if repro.auth is None:
            raise Exception("missing Auth: %s" % repro)

        with ssh_connect(repro.ip, repro.auth.private_key, call=True):
            pass

    def rdp(self, vm_id: str) -> None:
        vm_id = self._disambiguate(vm_id)
        repro = self.onefuzz.repro.get(vm_id)
        if repro.ip is None:
            raise Exception("missing IP: %s" % repro)
        if repro.auth is None:
            raise Exception("missing Auth: %s" % repro)

        RDP_PORT = 3389
        with rdp_connect(repro.ip, repro.auth.password, port=RDP_PORT):
            return


class DebugNode(Command):
    """Debug a specific node on a scaleset"""

    def rdp(self, machine_id: UUID_EXPANSION, duration: Optional[int] = 1) -> None:
        node = self.onefuzz.nodes.get(machine_id)
        if node.scaleset_id is None:
            raise Exception("node is not part of a scaleset")
        self.onefuzz.debug.scalesets.rdp(
            scaleset_id=node.scaleset_id, machine_id=node.machine_id, duration=duration
        )

    def ssh(self, machine_id: UUID_EXPANSION, duration: Optional[int] = 1) -> None:
        node = self.onefuzz.nodes.get(machine_id)
        if node.scaleset_id is None:
            raise Exception("node is not part of a scaleset")
        self.onefuzz.debug.scalesets.ssh(
            scaleset_id=node.scaleset_id, machine_id=node.machine_id, duration=duration
        )


class DebugScaleset(Command):
    """Debug tasks"""

    def _get_proxy_setup(
        self, scaleset_id: UUID, machine_id: UUID, port: int, duration: Optional[int]
    ) -> Tuple[bool, str, Optional[Tuple[str, int]]]:
        proxy = self.onefuzz.scaleset_proxy.create(
            scaleset_id, machine_id, port, duration=duration
        )
        if proxy.ip is None:
            return (False, "waiting on proxy ip", None)

        return (True, "waiting on proxy port", (proxy.ip, proxy.forward.src_port))

    def rdp(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        duration: Optional[int] = 1,
    ) -> None:
        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(
            scaleset_id, machine_id, include_auth=True
        )

        RDP_PORT = 3389
        setup = wait(
            lambda: self._get_proxy_setup(
                scaleset.scaleset_id, machine_id_expanded, RDP_PORT, duration
            )
        )
        if setup is None:
            raise Exception("no proxy for RDP port configured")

        if scaleset.auth is None:
            raise Exception("auth is not available for scaleset")

        ip, port = setup
        with rdp_connect(ip, scaleset.auth.password, port=port):
            return

    def ssh(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        duration: Optional[int] = 1,
        command: Optional[str] = None,
    ) -> None:
        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(
            scaleset_id, machine_id, include_auth=True
        )

        SSH_PORT = 22
        setup = wait(
            lambda: self._get_proxy_setup(
                scaleset.scaleset_id, machine_id_expanded, SSH_PORT, duration
            )
        )
        if setup is None:
            raise Exception("no proxy for SSH port configured")

        ip, port = setup

        if scaleset.auth is None:
            raise Exception("auth is not available for scaleset")

        with ssh_connect(
            ip, scaleset.auth.private_key, port=port, call=True, command=command
        ):
            return


class DebugTask(Command):
    """Debug a specific job"""

    def list_nodes(self, task_id: UUID_EXPANSION) -> Optional[List[NodeAssignment]]:
        task = self.onefuzz.tasks.get(task_id)
        return task.nodes

    def _get_node(
        self, task_id: UUID_EXPANSION, node_id: Optional[UUID]
    ) -> Tuple[UUID, UUID]:
        nodes = self.list_nodes(task_id)
        if not nodes:
            raise Exception("task is not currently executing on nodes")

        if node_id is not None:
            for node in nodes:
                if node.node_id == node_id and node.scaleset_id is not None:
                    return (node.scaleset_id, node.node_id)
            raise Exception("unable to find scaleset with node_id")

        for node in nodes:
            if node.scaleset_id:
                return (node.scaleset_id, node.node_id)

        raise Exception("unable to find scaleset node running on task")

    def ssh(
        self,
        task_id: UUID_EXPANSION,
        *,
        node_id: Optional[UUID] = None,
        duration: Optional[int] = 1,
    ) -> None:
        scaleset_id, node_id = self._get_node(task_id, node_id)
        return self.onefuzz.debug.scalesets.ssh(scaleset_id, node_id, duration=duration)

    def rdp(
        self,
        task_id: UUID_EXPANSION,
        *,
        node_id: Optional[UUID] = None,
        duration: Optional[int] = 1,
    ) -> None:
        scaleset_id, node_id = self._get_node(task_id, node_id)
        return self.onefuzz.debug.scalesets.rdp(scaleset_id, node_id, duration=duration)

    def libfuzzer_coverage(
        self,
        task_id: UUID_EXPANSION,
        timespan: str = DAY_TIMESPAN,
        limit: Optional[int] = None,
    ) -> Any:
        """
        Get the coverage for the specified task

        :param task_id value: Task ID
        :param str timespan: ISO 8601 duration format
        :param int limit: Limit the number of records returned
        """
        task = self.onefuzz.tasks.get(task_id)
        query = f"where customDimensions.task_id == '{task.task_id}'"
        return self.onefuzz.debug.logs._query_libfuzzer_coverage(query, timespan, limit)

    def libfuzzer_execs_sec(
        self,
        task_id: UUID_EXPANSION,
        timespan: str = DAY_TIMESPAN,
        limit: Optional[int] = None,
    ) -> Any:
        """
        Get the executions per second for the specified task

        :param task_id value: Task ID
        :param str timespan: ISO 8601 duration format
        :param int limit: Limit the number of records returned
        """
        task = self.onefuzz.tasks.get(task_id)
        query = f"where customDimensions.task_id == '{task.task_id}'"
        return self.onefuzz.debug.logs._query_libfuzzer_execs_sec(
            query, timespan, limit
        )


class DebugJobTask(Command):
    """Debug a task for a specific job"""

    def _get_task(self, job_id: UUID_EXPANSION, task_type: TaskType) -> UUID:
        for task in self.onefuzz.tasks.list(job_id=job_id):
            if task.config.task.type == task_type:
                return task.task_id

        raise Exception(
            "unable to find task type %s for job:%s" % (task_type.name, job_id)
        )

    def ssh(
        self,
        job_id: UUID_EXPANSION,
        task_type: TaskType,
        *,
        duration: Optional[int] = 1,
    ) -> None:
        """SSH into the first node running the specified task type in the job"""
        return self.onefuzz.debug.task.ssh(
            self._get_task(job_id, task_type), duration=duration
        )

    def rdp(
        self,
        job_id: UUID_EXPANSION,
        task_type: TaskType,
        *,
        duration: Optional[int] = 1,
    ) -> None:
        """RDP into the first node running the specified task type in the job"""
        return self.onefuzz.debug.task.rdp(
            self._get_task(job_id, task_type), duration=duration
        )


class DebugJob(Command):
    """Debug a specific Job"""

    def __init__(self, onefuzz: Any, logger: logging.Logger):
        super().__init__(onefuzz, logger)
        self.task = DebugJobTask(onefuzz, logger)

    def libfuzzer_coverage(
        self,
        job_id: UUID_EXPANSION,
        timespan: str = DAY_TIMESPAN,
        limit: Optional[int] = None,
    ) -> Any:
        """
        Get the coverage for the specified job

        :param job_id value: Job ID
        :param str timespan: ISO 8601 duration format
        :param int limit: Limit the number of records returned
        """
        job = self.onefuzz.jobs.get(job_id)
        query = f"where customDimensions.job_id == '{job.job_id}'"
        return self.onefuzz.debug.logs._query_libfuzzer_coverage(query, timespan, limit)

    def libfuzzer_execs_sec(
        self,
        job_id: UUID_EXPANSION,
        timespan: str = DAY_TIMESPAN,
        limit: Optional[int] = None,
    ) -> Any:
        """
        Get the executions per second for the specified job

        :param job_id value: Job ID
        :param str timespan: ISO 8601 duration format
        :param int limit: Limit the number of records returned
        """
        job = self.onefuzz.jobs.get(job_id)
        query = f"where customDimensions.job_id == '{job.job_id}'"
        return self.onefuzz.debug.logs._query_libfuzzer_execs_sec(
            query, timespan, limit
        )

    def download_files(self, job_id: UUID_EXPANSION, output: Directory) -> None:
        """Download the containers by container type for each task in the specified job"""

        self.onefuzz.jobs.containers.download(job_id, output=output)

    def rerun(
        self,
        job_id: UUID_EXPANSION,
        *,
        duration: Optional[int] = None,
        pool_name: Optional[PoolName] = None,
    ) -> Job:
        """rerun a given job"""

        existing_job = self.onefuzz.jobs.get(job_id)
        job_config = existing_job.config
        if duration is not None:
            job_config.duration = duration
        existing_tasks = self.onefuzz.jobs.tasks.list(existing_job.job_id)

        configs: Dict[UUID, TaskConfig] = {}
        todo: Set[UUID] = set()
        new_task_ids: Dict[UUID, UUID] = {}

        for task in existing_tasks:
            config = TaskConfig.parse_obj(json.loads(task.config.json()))
            if pool_name is not None and config.pool is not None:
                config.pool.pool_name = pool_name
            configs[task.task_id] = config
            todo.add(task.task_id)

        job = self.onefuzz.jobs.create_with_config(existing_job.config)

        while todo:
            added: Set[UUID] = set()

            for task_id in todo:
                config = configs[task_id]
                config.job_id = job.job_id
                if config.prereq_tasks:
                    if set(config.prereq_tasks).issubset(new_task_ids.keys()):
                        config.prereq_tasks = [
                            new_task_ids[x] for x in config.prereq_tasks
                        ]
                        task = self.onefuzz.tasks.create_with_config(config)
                        self.logger.info(
                            "created task: %s - %s",
                            task.task_id,
                            task.config.task.type.name,
                        )
                        new_task_ids[task_id] = task.task_id
                        added.add(task_id)
                else:
                    task = self.onefuzz.tasks.create_with_config(config)
                    self.logger.info(
                        "created task: %s - %s",
                        task.task_id,
                        task.config.task.type.name,
                    )
                    new_task_ids[task_id] = task.task_id
                    added.add(task_id)

            if added:
                todo -= added
            else:
                raise Exception(f"unable to resolve task prereqs for: {todo}")

        return job


class DebugLog(Command):
    def __init__(self, onefuzz: "Onefuzz", logger: logging.Logger):
        self.onefuzz = onefuzz
        self.logger = logger
        self._client: Optional[ApplicationInsightsDataClient] = None
        self._app_id: Optional[str] = None

    def _convert(self, raw_data: Any) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        results = {}
        for table in raw_data.tables:
            result = []
            for row in table.rows:
                converted = {
                    table.columns[x].name: y
                    for (x, y) in enumerate(row)
                    if y not in [None, ""]
                }
                if "customDimensions" in converted:
                    converted["customDimensions"] = json.loads(
                        converted["customDimensions"]
                    )
                result.append(converted)
            results[table.name] = result

        if list(results.keys()) == ["PrimaryResult"]:
            return results["PrimaryResult"]

        return results

    def query(
        self,
        log_query: str,
        *,
        timespan: Optional[str] = DAY_TIMESPAN,
        raw: bool = False,
    ) -> Any:
        """
        Perform an Application Insights query

        Queries should be well formed Kusto Queries.
        Ref https://docs.microsoft.com/en-us/azure/data-explorer/kql-quick-reference

        :param str log_query: Query to send to Application Insights
        :param str timespan: ISO 8601 duration format
        :param bool raw: Do not simplify the data result
        """
        if self._app_id is None:
            self._app_id = self.onefuzz.info.get().insights_appid
        if self._app_id is None:
            raise Exception("instance does not have an insights_appid")
        if self._client is None:

            creds, _ = get_azure_cli_credentials(
                resource="https://api.applicationinsights.io"
            )
            self._client = ApplicationInsightsDataClient(creds)

        self.logger.debug("query: %s", log_query)
        raw_data = self._client.query.execute(
            self._app_id, body=QueryBody(query=log_query, timespan=timespan)
        )
        if "error" in raw_data.additional_properties:
            raise Exception(
                "Error performing query: %s" % raw_data.additional_properties["error"]
            )
        if raw:
            return raw_data
        return self._convert(raw_data)

    def _query_parts(
        self, parts: List[str], *, timespan: Optional[str] = None, raw: bool = False
    ) -> Any:
        log_query = " | ".join(parts)
        return self.query(log_query, timespan=timespan, raw=raw)

    def _build_keyword_query(
        self, value: str, limit: Optional[int] = None, desc: bool = True
    ) -> List[str]:
        # See https://docs.microsoft.com/en-us/azure/data-explorer/kql-quick-reference

        components = ["union isfuzzy=true exceptions, traces, customEvents"]
        value = value.strip()
        keywords = ['* has "%s"' % (x.replace('"', '\\"')) for x in value.split(" ")]
        if keywords:
            components.append("where " + " and ".join(keywords))
        order = "desc" if desc else "asc"
        if limit:
            components.append(f"take {limit}")
        components.append(f"order by timestamp {order}")
        return components

    def keyword(
        self,
        value: str,
        *,
        timespan: Optional[str] = DAY_TIMESPAN,
        limit: Optional[int] = None,
        raw: bool = False,
    ) -> Any:
        """
        Perform an Application Insights keyword query akin to "Transaction Search"

        :param str value: Keyword to query Application Insights
        :param str timespan: ISO 8601 duration format
        :param int limit: Limit the number of records returned
        :param bool raw: Do not simplify the data result
        """

        components = self._build_keyword_query(value, limit=limit)

        return self._query_parts(components, timespan=timespan, raw=raw)

    def tail(
        self,
        value: str,
        *,
        limit: int = 1000,
        indent: Optional[int] = None,
        filter: Optional[str] = "[message, name, customDimensions]",
        timespan: Optional[str] = HOUR_TIMESPAN,
    ) -> None:
        """
        Perform an Application Insights keyword query akin to "Transaction Search"

        :param str value: Keyword to query Application Insights
        :param str indent: Specify indent for JSON printing
        :param str limit: Limit the number of records to return in each query
        :param str filter: JMESPath filter for streaming results
        """

        expression = None
        if filter:
            expression = jmespath.compile(filter)

        base_query = self._build_keyword_query(value, limit=limit, desc=False)

        last_seen: Optional[str] = None
        wait = DEFAULT_TAIL_DELAY

        while True:
            query = base_query.copy()
            if last_seen is not None:
                query.append(f'where timestamp > datetime("{last_seen}")')
            results = self._query_parts(query, timespan=timespan)

            if results:
                last_seen = results[-1]["timestamp"]
                for entry in results:
                    if expression is not None:
                        entry = expression.search(entry)
                    if entry:
                        print(json.dumps(entry, indent=indent, sort_keys=True))
                wait = DEFAULT_TAIL_DELAY
            else:
                self.onefuzz.logger.debug("waiting %f seconds", wait)

                time.sleep(wait)
                if wait < 60:
                    wait *= 1.5

    def _query_libfuzzer_coverage(
        self, query: str, timespan: str, limit: Optional[int] = None
    ) -> Any:
        project_fields = [
            "rate=customDimensions.rate",
            "covered=customDimensions.covered",
            "features=customDimensions.features",
            "timestamp",
        ]

        query_parts = [
            "customEvents",
            "where name == 'coverage_data'",
            query,
            "order by timestamp desc",
            f"project {','.join(project_fields)}",
        ]

        if limit:
            query_parts.append(f"take {limit}")

        return self.onefuzz.debug.logs._query_parts(query_parts, timespan=timespan)

    def _query_libfuzzer_execs_sec(
        self,
        query: str,
        timespan: str,
        limit: Optional[int] = None,
    ) -> Any:
        project_fields = [
            "machine_id=customDimensions.machine_id",
            "worker_id=customDimensions.worker_id",
            "execs_sec=customDimensions.execs_sec",
            "timestamp",
        ]

        query_parts = [
            "customEvents",
            "where name == 'runtime_stats'",
            query,
            "where customDimensions.execs_sec > 0",
            "order by timestamp desc",
            f"project {','.join(project_fields)}",
        ]
        if limit:
            query_parts.append(f"take {limit}")

        return self.onefuzz.debug.logs._query_parts(query_parts, timespan=timespan)


class DebugNotification(Command):
    """Debug notification integrations"""

    def _get_container(
        self, task: Task, container_type: ContainerType
    ) -> Optional[Container]:
        for container in task.config.containers:
            if container.type == container_type:
                return container.name
        return None

    def _get_storage_account(self, container_name: Container) -> str:
        sas_url = self.onefuzz.containers.get(container_name).sas_url
        _, netloc, _, _, _, _ = urlparse(sas_url)
        return netloc.split(".")[0]

    def job(
        self,
        job_id: str,
        *,
        report_container_type: ContainerType = ContainerType.unique_reports,
        crash_name: str = "fake-crash-sample",
    ) -> None:
        """Inject a report into the first crash reporting task in the specified job"""

        tasks = self.onefuzz.tasks.list(job_id=job_id, state=[])
        for task in tasks:
            if task.config.task.type in [
                TaskType.libfuzzer_crash_report,
                TaskType.generic_crash_report,
            ]:
                self.task(
                    str(task.task_id),
                    report_container_type=report_container_type,
                    crash_name=crash_name,
                )
                return

        raise Exception("no crash reporting tasks configured")

    def task(
        self,
        task_id: str,
        *,
        report_container_type: ContainerType = ContainerType.unique_reports,
        crash_name: str = "fake-crash-sample",
    ) -> None:
        """Inject a report into the specified crash reporting task"""

        task = self.onefuzz.tasks.get(task_id)
        crashes = self._get_container(task, ContainerType.crashes)
        reports = self._get_container(task, report_container_type)

        if crashes is None:
            raise Exception("task does not have a crashes container")

        if reports is None:
            raise Exception(
                "task does not have a %s container" % report_container_type.name
            )

        with tempfile.TemporaryDirectory() as tempdir:
            file_path = os.path.join(tempdir, crash_name)
            with open(file_path, "w") as handle:
                handle.write("")
            self.onefuzz.containers.files.upload_file(crashes, file_path, crash_name)

        report = Report(
            input_blob=BlobRef(
                account=self._get_storage_account(crashes),
                container=crashes,
                name=crash_name,
            ),
            executable=task.config.task.target_exe,
            crash_type="fake crash report",
            crash_site="fake crash site",
            call_stack=["#0 fake", "#1 call", "#2 stack"],
            call_stack_sha256=ZERO_SHA256,
            input_sha256=EMPTY_SHA256,
            asan_log="fake asan log",
            task_id=task_id,
            job_id=task.job_id,
        )

        with tempfile.TemporaryDirectory() as tempdir:
            file_path = os.path.join(tempdir, "report.json")
            with open(file_path, "w") as handle:
                handle.write(report.json())

            self.onefuzz.containers.files.upload_file(
                reports, file_path, crash_name + ".json"
            )


class Debug(Command):
    """Debug running jobs"""

    def __init__(self, onefuzz: Any, logger: logging.Logger):
        super().__init__(onefuzz, logger)
        self.scalesets = DebugScaleset(onefuzz, logger)
        self.repro = DebugRepro(onefuzz, logger)
        self.job = DebugJob(onefuzz, logger)
        self.notification = DebugNotification(onefuzz, logger)
        self.task = DebugTask(onefuzz, logger)
        self.logs = DebugLog(onefuzz, logger)
        self.node = DebugNode(onefuzz, logger)



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
from inspect import Parameter, signature
from typing import Any, Dict, List, Optional

from onefuzztypes.enums import ContainerType, UserFieldType
from onefuzztypes.job_templates import JobTemplateConfig, JobTemplateRequestParameters
from onefuzztypes.models import Job
from onefuzztypes.primitives import Directory, File

from ..api import PreviewFeature
from .handlers import TemplateSubmitHandler

LOGGER = logging.getLogger("job-templates")

TYPES = {
    UserFieldType.Str: str,
    UserFieldType.Int: int,
    UserFieldType.ListStr: List[str],
    UserFieldType.DictStr: Dict[str, str],
    UserFieldType.Bool: bool,
}

NAMES = {
    UserFieldType.Str: "str",
    UserFieldType.Int: "int",
    UserFieldType.ListStr: "list",
    UserFieldType.DictStr: "dict",
    UserFieldType.Bool: "bool",
}


def container_type_name(container_type: ContainerType) -> str:
    return container_type.name + "_dir"


def config_to_params(config: JobTemplateConfig) -> List[Parameter]:
    params: List[Parameter] = []

    for entry in config.user_fields:
        default = entry.default if entry.default is not None else Parameter.empty
        annotation: Any = None

        if entry.name == "target_exe":
            annotation = File
        else:
            annotation = TYPES[entry.type]

        is_optional = entry.default is None and entry.required is False
        if is_optional:
            annotation = Optional[annotation]

        param = Parameter(
            entry.name,
            Parameter.KEYWORD_ONLY,
            annotation=annotation,
            default=default,
        )
        params.append(param)

    for container in config.containers:
        if container not in ContainerType.user_config():
            continue

        param = Parameter(
            container_type_name(container),
            Parameter.KEYWORD_ONLY,
            annotation=Optional[Directory],
            default=Parameter.empty,
        )
        params.append(param)

    if config.containers:
        param = Parameter(
            "container_names",
            Parameter.KEYWORD_ONLY,
            annotation=Optional[Dict[ContainerType, str]],
            default=None,
        )
        params.append(param)

    param = Parameter(
        "parameters",
        Parameter.KEYWORD_ONLY,
        annotation=Optional[JobTemplateRequestParameters],
        default=Parameter.empty,
    )
    params.append(param)

    return params


def build_template_doc(config: JobTemplateConfig) -> str:
    docs = [
        f"Launch '{config.name}' job",
        "",
    ]

    for entry in config.user_fields:
        line = f":param {NAMES[entry.type]} {entry.name}: {entry.help}"
        docs.append(line)

    for container in config.containers:
        if container not in ContainerType.user_config():
            continue
        line = f":param Directory {container_type_name(container)}: Local path to the {container.name} directory"
        docs.append(line)

    if config.containers:
        line = ":param dict container_names: custom container names (eg: setup=my-setup-container)"
        docs.append(line)

    return "\n".join(docs)


def build_template_func(config: JobTemplateConfig) -> Any:
    def func(
        self: TemplateSubmitHandler,
        *,
        wait_for_running: bool = False,
        **kwargs: Any,
    ) -> Job:
        self.onefuzz._warn_preview(PreviewFeature.job_templates)
        return self._execute(
            config,
            kwargs,
            wait_for_running=wait_for_running,
        )

    sig = signature(func)
    params = [
        sig.parameters["self"],
        sig.parameters["wait_for_running"],
    ] + config_to_params(config)
    sig = sig.replace(parameters=tuple(params))
    func.__signature__ = sig  # type: ignore

    func.__doc__ = build_template_doc(config)

    return func



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import configparser
import glob
import os
import subprocess  # nosec
from typing import Dict, List, Optional, Tuple

from onefuzztypes.enums import OS, ContainerType, TaskDebugFlag
from onefuzztypes.models import NotificationConfig
from onefuzztypes.primitives import File, PoolName

from onefuzz.api import Command
from onefuzz.backend import container_file_path

from . import JobHelper

VM_SKU = "Standard_D2s_v3"
VM_COUNT = 1


class OssFuzz(Command):
    """OssFuzz style jobs"""

    def _containers(self, project: str, build: str, platform: OS) -> Dict[str, str]:
        guid = self.onefuzz.utils.namespaced_guid(
            project, build=build, platform=platform.name
        ).hex

        names = {
            "build": "oss-build-%s" % guid,
            "base": "oss-base-%s" % guid,
        }
        return names

    @classmethod
    def _options(_cls, filename: File) -> Tuple[Dict[str, str], List[str]]:
        target_env: Dict[str, str] = {}
        target_options: List[str] = []

        if not os.path.exists(filename):
            return target_env, target_options

        config = configparser.ConfigParser()
        with open(filename, "r") as handle:
            config.read_file(handle)

            if config.has_section("env"):
                target_env.update({x[0].upper(): x[1] for x in config.items("env")})

            if config.has_section("libfuzzer"):
                for arg, value in config.items("libfuzzer"):
                    if arg == "dict":
                        value = "setup/%s" % value
                    target_options.append("-%s=%s" % (arg, value))

        return target_env, target_options

    @classmethod
    def _owners(_cls, filename: File) -> Dict[str, str]:
        tags = {}

        if os.path.exists(filename):
            with open(filename, "r") as handle:
                owners = handle.read()
            owner, ado_path = owners.split(",")
            tags["EMAIL"] = owner
            tags["ADO_PATH"] = ado_path

        return tags

    def _copy_all(self, src_sas: str, dst_sas: str) -> None:
        cmd = [
            "azcopy",
            "sync",
            src_sas,
            dst_sas,
        ]
        self.logger.info("copying base setup")
        # security note: the source and destination container sas URLS are
        # considerd trusted from the service
        subprocess.check_output(cmd)  # nosec

    def _copy_exe(self, src_sas: str, dst_sas: str, target_exe: File) -> None:
        files: List[File] = [target_exe]
        pdb_path = os.path.splitext(target_exe)[0] + ".pdb"
        if os.path.exists(str(pdb_path)):
            files.append(File(pdb_path))

        for path in files:
            filename = os.path.basename(path)
            src_url = container_file_path(src_sas, filename)
            dst_url = container_file_path(dst_sas, filename)

            cmd = [
                "azcopy",
                "copy",
                src_url,
                dst_url,
            ]
            self.logger.info("uploading %s", path)
            # security note: the source and destination container sas URLS
            # are considerd trusted from the service
            subprocess.check_output(cmd)  # nosec

    def libfuzzer(
        self,
        project: str,
        build: str,
        pool_name: PoolName,
        duration: int = 24,
        tags: Optional[Dict[str, str]] = None,
        dryrun: bool = False,
        max_target_count: int = 20,
        sync_inputs: bool = False,
        notification_config: Optional[NotificationConfig] = None,
        debug: Optional[List[TaskDebugFlag]] = None,
        ensemble_sync_delay: Optional[int] = None,
    ) -> None:
        """
        OssFuzz style libfuzzer jobs

        :param bool ensemble_sync_delay: Specify duration between
            syncing inputs during ensemble fuzzing (0 to disable).
        """

        fuzzers = sorted(glob.glob("*fuzzer"))
        if fuzzers:
            platform = OS.linux
        else:
            platform = OS.windows
            fuzzers = sorted(glob.glob("*fuzzer.exe"))

        if dryrun:
            return

        containers = self._containers(project, build, platform)
        container_sas = {}
        for name in containers:
            self.logger.info("creating container: %s", name)
            sas_url = self.onefuzz.containers.create(containers[name]).sas_url
            container_sas[name] = sas_url

        self.logger.info("uploading build artifacts")

        # security note: the container sas is considered trusted from the
        # service
        subprocess.check_output(  # nosec
            [
                "azcopy",
                "sync",
                ".",
                container_sas["build"],
                "--exclude-pattern",
                "*fuzzer_seed_corpus.zip",
            ]
        )
        subprocess.check_output(  # nosec
            [
                "azcopy",
                "sync",
                ".",
                container_sas["base"],
                '--include-pattern="*.so;*.dll;*.sh;*.ps1',
            ]
        )

        if max_target_count:
            fuzzers = fuzzers[:max_target_count]

        base_helper = JobHelper(
            self.onefuzz,
            self.logger,
            project,
            build,
            "base",
            duration,
            pool_name=pool_name,
            target_exe=File(fuzzers[0]),
        )
        base_helper.platform = platform

        helpers = []
        for fuzzer in [File(x) for x in fuzzers]:
            fuzzer_name = fuzzer.replace(".exe", "").replace("_fuzzer", "")
            self.logger.info("creating tasks for %s", fuzzer)
            self.onefuzz.template.libfuzzer._check_is_libfuzzer(fuzzer)
            helper = JobHelper(
                self.onefuzz,
                self.logger,
                project,
                fuzzer_name,
                build,
                duration,
                job=base_helper.job,
                pool_name=pool_name,
                target_exe=fuzzer,
            )
            helper.platform = platform
            helper.add_tags(tags)
            helper.platform = base_helper.platform
            helper.job = base_helper.job
            helper.define_containers(
                ContainerType.setup,
                ContainerType.inputs,
                ContainerType.crashes,
                ContainerType.reports,
                ContainerType.unique_reports,
                ContainerType.no_repro,
                ContainerType.coverage,
            )
            helper.create_containers()
            helper.setup_notifications(notification_config)

            dst_sas = self.onefuzz.containers.get(
                helper.containers[ContainerType.setup]
            ).sas_url
            self._copy_exe(container_sas["build"], dst_sas, File(fuzzer))
            self._copy_all(container_sas["base"], dst_sas)

            zip_name = fuzzer.replace(".exe", "").replace("_fuzzer", "_seed_corpus.zip")

            if os.path.exists(zip_name) and sync_inputs:
                self.logger.info("uploading seeds")
                helper.upload_inputs_zip(File(zip_name))

            owners_path = File("%s.msowners" % fuzzer.replace(".exe", ""))
            options_path = File("%s.options" % fuzzer.replace(".exe", ""))

            target_env, target_options = self._options(options_path)
            helper.add_tags(self._owners(owners_path))

            # All fuzzers are copied to the setup container root.
            #
            # Cast because `glob()` returns `str`.
            fuzzer_blob_name = helper.setup_relative_blob_name(fuzzer, None)

            self.onefuzz.template.libfuzzer._create_tasks(
                job=base_helper.job,
                containers=helper.containers,
                pool_name=pool_name,
                target_exe=fuzzer_blob_name,
                vm_count=VM_COUNT,
                duration=duration,
                target_options=target_options,
                target_env=target_env,
                tags=helper.tags,
                debug=debug,
                ensemble_sync_delay=ensemble_sync_delay,
            )
            helpers.append(helper)
        base_helper.wait()



import os
import shutil
import subprocess  # nosec


def azcopy_sync(src: str, dst: str) -> None:
    """Expose azcopy for uploading/downloading files"""

    azcopy = os.environ.get("AZCOPY") or shutil.which("azcopy")
    if not azcopy:
        raise Exception(
            "unable to find 'azcopy' in path or AZCOPY environment variable"
        )

    # security note: callers need to understand the src/dst for this.
    subprocess.check_output([azcopy, "sync", src, dst, "--recursive=true"])  # nosec



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from collections import defaultdict
from typing import DefaultDict, List, Optional, Set
from uuid import UUID

from onefuzztypes.enums import ContainerType, JobState
from onefuzztypes.primitives import Container

from ..api import UUID_EXPANSION, Command
from .cache import JobFilter
from .raw import raw
from .top import Top


def short(value: UUID) -> str:
    return str(value).split("-")[0]


class Status(Command):
    """Monitor status of Onefuzz Instance"""

    def project(
        self,
        *,
        project: str,
        name: str,
        build: Optional[str] = None,
        all_jobs: Optional[bool] = False,
    ) -> None:
        job_state = JobState.available()
        if all_jobs:
            job_state = [x for x in JobState]

        for job in self.onefuzz.jobs.list(job_state=job_state):
            if job.config.project != project:
                continue
            if job.config.name != name:
                continue
            if build is not None and job.config.build != build:
                continue

            self.job(job.job_id)

    def job(self, job_id: UUID_EXPANSION) -> None:
        job = self.onefuzz.jobs.get(job_id)

        print(f"job: {job.job_id}")
        print(
            f"project:{job.config.project} name:{job.config.name} build:{job.config.build}"
        )
        print("\ntasks:")

        tasks = self.onefuzz.tasks.list(job_id=job.job_id, state=[])
        with_errors = []
        for task in tasks:
            if task.error:
                with_errors.append(task)
                continue
            print(
                f"{short(task.task_id)} "
                f"target:{task.config.task.target_exe} "
                f"state:{task.state.name} "
                f"type:{task.config.task.type.name}"
            )

        if with_errors:
            print("\ntasks with errors:")
            entries = []
            for task in with_errors:
                if not task.error:
                    continue

                errors = "\n".join([x.strip() for x in task.error.errors])

                message = (
                    f"{short(task.task_id)} type:{task.config.task.type.name} "
                    f"target:{task.config.task.target_exe}"
                    f"\nerror:\n{errors}"
                )
                entries.append(message)
            print("\n\n".join(entries))

        containers: DefaultDict[ContainerType, Set[Container]] = defaultdict(set)
        for task in tasks:
            for container in task.config.containers:
                if container.type not in containers:
                    containers[container.type] = set()
                containers[container.type].add(container.name)

        print("\ncontainers:")
        for container_type in containers:
            for container_name in containers[container_type]:
                try:
                    count = len(
                        self.onefuzz.containers.files.list(container_name).files
                    )
                    print(
                        f"{container_type.name:<15} count:{count:<5} name:{container_name}"
                    )
                except Exception:
                    print(
                        f"{container_type.name:<15} INACCESSIBLE CONTAINER name:{container_name}"
                    )

    def raw(self) -> None:
        """Raw status update stream"""
        raw(self.onefuzz, self.logger)

    def top(
        self,
        *,
        job_id: Optional[List[UUID]] = None,
        project: Optional[List[str]] = None,
        name: Optional[List[str]] = None,
    ) -> None:
        """Onefuzz Top"""
        job_filter = JobFilter(job_id=job_id, project=project, name=name)
        top = Top(self.onefuzz, self.logger, job_filter)
        top.run()



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json
import logging
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from uuid import UUID

from onefuzztypes.enums import ContainerType, JobState, NodeState, TaskState, TaskType
from onefuzztypes.events import (
    EventCrashReported,
    EventFileAdded,
    EventJobCreated,
    EventJobStopped,
    EventNodeCreated,
    EventNodeDeleted,
    EventNodeStateUpdated,
    EventPoolCreated,
    EventPoolDeleted,
    EventTaskCreated,
    EventTaskFailed,
    EventTaskStateUpdated,
    EventTaskStopped,
    EventType,
    parse_event_message,
)
from onefuzztypes.models import (
    Job,
    JobConfig,
    Node,
    Pool,
    Task,
    TaskContainers,
    UserInfo,
)
from onefuzztypes.primitives import Container, PoolName
from pydantic import BaseModel

MESSAGE = Tuple[datetime, EventType, str]

MINUTES = 60
HOURS = 60 * MINUTES
DAYS = 24 * HOURS


# status-top only representation of a Node
class MiniNode(BaseModel):
    machine_id: UUID
    pool_name: PoolName
    state: NodeState


# status-top only representation of a Job
class MiniJob(BaseModel):
    job_id: UUID
    config: JobConfig
    state: Optional[JobState]
    user_info: Optional[UserInfo]


# status-top only representation of a Task
class MiniTask(BaseModel):
    job_id: UUID
    task_id: UUID
    type: TaskType
    target: str
    state: TaskState
    pool: str
    end_time: Optional[datetime]
    containers: List[TaskContainers]


def fmt(data: Any) -> Any:
    if data is None:
        return ""
    if isinstance(data, int):
        return str(data)
    if isinstance(data, str):
        return data
    if isinstance(data, UUID):
        return str(data)[:8]
    if isinstance(data, list):
        return [fmt(x) for x in data]
    if isinstance(data, datetime):
        return data.strftime("%H:%M:%S")
    if isinstance(data, tuple):
        return tuple([fmt(x) for x in data])
    if isinstance(data, Enum):
        return data.name
    if isinstance(data, dict):
        return " ".join(
            sorted(
                [
                    "{}:{}".format(fmt(x).title().replace("_", " "), fmt(y))
                    for (x, y) in data.items()
                ]
            )
        )
    raise NotImplementedError(type(data))


class JobFilter(BaseModel):
    job_id: Optional[List[UUID]]
    project: Optional[List[str]]
    name: Optional[List[str]]


class TopCache:
    JOB_FIELDS = ["Job", "Name", "User", "Files"]
    TASK_FIELDS = [
        "Job",
        "Task",
        "State",
        "Type",
        "Target",
        "Files",
        "Pool",
        "End time",
    ]
    POOL_FIELDS = ["Name", "OS", "Arch", "Nodes"]

    def __init__(
        self,
        onefuzz: "Onefuzz",
        job_filters: JobFilter,
    ):
        self.onefuzz = onefuzz
        self.job_filters = job_filters
        self.tasks: Dict[UUID, MiniTask] = {}
        self.jobs: Dict[UUID, MiniJob] = {}
        self.files: Dict[Container, Set[str]] = {}
        self.pools: Dict[str, EventPoolCreated] = {}
        self.nodes: Dict[UUID, MiniNode] = {}

        self.messages: List[MESSAGE] = []
        endpoint = onefuzz._backend.config.endpoint
        if not endpoint:
            raise Exception("endpoint is not set")
        self.endpoint: str = endpoint
        self.last_update = datetime.now()

    def add_container(self, name: Container) -> None:
        if name in self.files:
            return
        try:
            files = self.onefuzz.containers.files.list(name)
        except Exception:
            return

        self.add_files_set(name, set(files.files))

    def add_message(self, message_obj: Any) -> None:
        message = parse_event_message(message_obj)

        event = message.event
        if isinstance(event, EventPoolCreated):
            self.pool_created(event)
        elif isinstance(event, EventPoolDeleted):
            self.pool_deleted(event)
        elif isinstance(event, EventTaskCreated):
            self.task_created(event)
        elif isinstance(event, EventTaskStopped):
            self.task_stopped(event)
        elif isinstance(event, EventTaskFailed):
            self.task_failed(event)
        elif isinstance(event, EventTaskStateUpdated):
            self.task_state_updated(event)
        elif isinstance(event, EventJobCreated):
            self.job_created(event)
        elif isinstance(event, EventJobStopped):
            self.job_stopped(event)
        elif isinstance(event, EventNodeStateUpdated):
            self.node_state_updated(event)
        elif isinstance(event, EventNodeCreated):
            self.node_created(event)
        elif isinstance(event, EventNodeDeleted):
            self.node_deleted(event)
        elif isinstance(event, (EventCrashReported, EventFileAdded)):
            self.file_added(event)

        self.last_update = datetime.now()
        messages = [x for x in self.messages][-99:]
        messages += [
            (
                datetime.now(),
                message.event_type,
                json.dumps(message_obj, sort_keys=True),
            )
        ]
        self.messages = messages

    def file_added(self, event: Union[EventFileAdded, EventCrashReported]) -> None:
        if event.container in self.files:
            files = self.files[event.container]
        else:
            files = set()
        files.update(set([event.filename]))
        self.files[event.container] = files

    def add_files_set(self, container: Container, new_files: Set[str]) -> None:
        if container in self.files:
            files = self.files[container]
        else:
            files = set()
        files.update(new_files)
        self.files[container] = files

    def add_node(self, node: Node) -> None:
        self.nodes[node.machine_id] = MiniNode(
            machine_id=node.machine_id, state=node.state, pool_name=node.pool_name
        )

    def add_job(self, job: Job) -> MiniJob:
        mini_job = MiniJob(
            job_id=job.job_id,
            config=job.config,
            state=job.state,
            user_info=job.user_info,
        )
        self.jobs[job.job_id] = mini_job
        return mini_job

    def job_created(
        self,
        job: EventJobCreated,
    ) -> None:
        self.jobs[job.job_id] = MiniJob(
            job_id=job.job_id, config=job.config, user_info=job.user_info
        )

    def add_pool(self, pool: Pool) -> None:
        self.pool_created(
            EventPoolCreated(
                pool_name=pool.name,
                os=pool.os,
                arch=pool.arch,
                managed=pool.managed,
            )
        )

    def pool_created(
        self,
        pool: EventPoolCreated,
    ) -> None:
        self.pools[pool.pool_name] = pool

    def pool_deleted(self, pool: EventPoolDeleted) -> None:
        if pool.pool_name in self.pools:
            del self.pools[pool.pool_name]

    def render_pools(self) -> List:
        results = []
        for pool in self.pools.values():
            nodes = {}
            for node in self.nodes.values():
                if node.pool_name != pool.pool_name:
                    continue
                if node.state not in nodes:
                    nodes[node.state] = 0
                nodes[node.state] += 1
            entry = (pool.pool_name, pool.os, pool.arch, nodes or "None")
            results.append(entry)
        return results

    def node_created(self, node: EventNodeCreated) -> None:
        self.nodes[node.machine_id] = MiniNode(
            machine_id=node.machine_id, pool_name=node.pool_name, state=NodeState.init
        )

    def node_state_updated(self, node: EventNodeStateUpdated) -> None:
        self.nodes[node.machine_id] = MiniNode(
            machine_id=node.machine_id, pool_name=node.pool_name, state=node.state
        )

    def node_deleted(self, node: EventNodeDeleted) -> None:
        if node.machine_id in self.nodes:
            del self.nodes[node.machine_id]

    def add_task(self, task: Task) -> None:
        self.tasks[task.task_id] = MiniTask(
            job_id=task.job_id,
            task_id=task.task_id,
            type=task.config.task.type,
            pool=task.config.pool.pool_name if task.config.pool else "",
            state=task.state,
            target=(task.config.task.target_exe or "").replace("setup/", "", 0),
            containers=task.config.containers,
            end_time=task.end_time,
        )

    def task_created(self, event: EventTaskCreated) -> None:
        self.tasks[event.task_id] = MiniTask(
            job_id=event.job_id,
            task_id=event.task_id,
            type=event.config.task.type,
            pool=event.config.pool.pool_name if event.config.pool else "",
            target=(event.config.task.target_exe or "").replace("setup/", "", 0),
            containers=event.config.containers,
            state=TaskState.init,
        )

    def task_state_updated(self, event: EventTaskStateUpdated) -> None:
        if event.task_id in self.tasks:
            task = self.tasks[event.task_id]
            task.state = event.state
            task.end_time = event.end_time
            self.tasks[event.task_id] = task

    def task_stopped(self, event: EventTaskStopped) -> None:
        if event.task_id in self.tasks:
            del self.tasks[event.task_id]

    def task_failed(self, event: EventTaskFailed) -> None:
        if event.task_id in self.tasks:
            del self.tasks[event.task_id]

    def render_tasks(self) -> List:
        results = []
        for task in self.tasks.values():
            job_entry = self.jobs.get(task.job_id)
            if job_entry:
                if not self.should_render_job(job_entry):
                    continue

            files = self.get_file_counts([task])

            entry = (
                task.job_id,
                task.task_id,
                task.state,
                task.type.name,
                task.target,
                files,
                task.pool,
                task.end_time,
            )
            results.append(entry)
        return results

    def add_job_if_missing(self, job_id: UUID) -> None:
        if job_id in self.jobs:
            return
        job = self.onefuzz.jobs.get(job_id)
        self.add_job(job)

    def should_render_job(self, job: MiniJob) -> bool:
        if self.job_filters.job_id is not None:
            if job.job_id not in self.job_filters.job_id:
                logging.info("skipping:%s", job)
                return False

        if self.job_filters.project is not None:
            if job.config.project not in self.job_filters.project:
                logging.info("skipping:%s", job)
                return False

        if self.job_filters.name is not None:
            if job.config.name not in self.job_filters.name:
                logging.info("skipping:%s", job)
                return False

        return True

    def job_stopped(self, event: EventJobStopped) -> None:
        if event.job_id in self.jobs:
            del self.jobs[event.job_id]

        to_remove = [x.task_id for x in self.tasks.values() if x.job_id == event.job_id]

        for task_id in to_remove:
            del self.tasks[task_id]

    def render_jobs(self) -> List[Tuple]:
        results: List[Tuple] = []

        for job in self.jobs.values():
            if not self.should_render_job(job):
                continue

            files = self.get_file_counts(self.get_tasks(job.job_id), merge_inputs=True)

            for to_remove in [ContainerType.coverage, ContainerType.setup]:
                if to_remove in files:
                    del files[to_remove]

            entry = (
                job.job_id,
                "%s:%s:%s" % (job.config.project, job.config.name, job.config.build),
                job.user_info.upn if job.user_info else "",
                files,
            )
            results.append(entry)

        return results

    def get_file_counts(
        self, tasks: List[MiniTask], merge_inputs: bool = False
    ) -> Dict[ContainerType, int]:
        results: Dict[ContainerType, Dict[str, int]] = {}
        for task in tasks:
            for container in task.containers:
                if container.name not in self.files:
                    continue
                if merge_inputs and container.type == ContainerType.readonly_inputs:
                    container.type = ContainerType.inputs
                if container.type not in results:
                    results[container.type] = {}
                results[container.type][container.name] = len(
                    self.files[container.name]
                )

        results_merged = {}
        for k, v in results.items():
            value = sum(v.values())
            if value:
                results_merged[k] = value

        return results_merged

    def get_tasks(self, job_id: UUID) -> List[MiniTask]:
        result = []
        for task in self.tasks.values():
            if task.job_id == job_id:
                result.append(task)
        return result


from onefuzz.api import Onefuzz  # noqa: E402



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import datetime
import json
import logging
import os
from typing import Dict, List, Optional

from onefuzztypes.job_templates import JobTemplateConfig
from pydantic import BaseModel, Field

from ..backend import ONEFUZZ_BASE_PATH

TEMPLATE_CACHE = os.path.expanduser(os.path.join(ONEFUZZ_BASE_PATH, "templates.json"))


class EndpointCache(BaseModel):
    timestamp: datetime.datetime
    configs: List[JobTemplateConfig]


class CachedTemplates(BaseModel):
    entries: Dict[str, EndpointCache] = Field(default_factory=dict)

    @classmethod
    def add(cls, endpoint: str, configs: List[JobTemplateConfig]) -> None:
        cache = cls.load()
        cache.entries[endpoint] = EndpointCache(
            timestamp=datetime.datetime.utcnow(), configs=configs
        )
        cache.save()

    @classmethod
    def get(cls, endpoint: str) -> Optional[EndpointCache]:
        cache = cls.load()
        return cache.entries.get(endpoint)

    @classmethod
    def load(cls) -> "CachedTemplates":
        if not os.path.exists(TEMPLATE_CACHE):
            entry = cls()
            entry.save()
            return entry

        try:
            with open(TEMPLATE_CACHE, "r") as handle:
                raw = json.load(handle)

            return cls.parse_obj(raw)
        except Exception as err:
            logging.warning("unable to load template cache: %s", err)
            entry = cls()
            entry.save()
            return entry

    def save(self) -> None:
        with open(TEMPLATE_CACHE, "w") as handle:
            handle.write(self.json())



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import json
import logging
import os
import shutil
import subprocess  # nosec
import sys
import tempfile
from typing import Optional

from onefuzztypes.enums import OS, ContainerType, TaskType
from onefuzztypes.models import NotificationConfig
from onefuzztypes.primitives import Container

from onefuzz.api import Onefuzz
from onefuzz.templates import JobHelper

AZCOPY_PATH = os.environ.get("AZCOPY") or shutil.which("azcopy")
if not AZCOPY_PATH:
    raise Exception("unable to find 'azcopy' in path or AZCOPY environment variable")

FUZZER_NAME = "domato"
FUZZER_URL = "https://github.com/googleprojectzero/domato.git"

FUZZER_NAME = "domato"
FUZZER_URL = "https://github.com/googleprojectzero/domato.git"

PY_URL = "https://www.python.org/ftp/python/3.7.8/python-3.7.8-amd64.exe"
PY_INSTALLER_ARGS = "/passive InstallAllUsers=1 AssociateFiles=1 PrependPath=1"
PY_INSTALLER = """
Invoke-WebRequest -Uri "%s" -OutFile "c:/python-install.exe"
Start-Process "C:\\python-install.exe"  -ArgumentList "%s" -Wait
""" % (
    PY_URL,
    PY_INSTALLER_ARGS,
)


def add_setup_script(of: Onefuzz, container: Container) -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        setup = os.path.join(tmpdir, "setup.ps1")
        with open(setup, "w") as handle:
            handle.write(PY_INSTALLER)
        of.containers.files.upload_file(container, setup)


def upload_to_fuzzer_container(of: Onefuzz, fuzzer_name: str, fuzzer_url: str) -> None:
    fuzzer_sas = of.containers.create(Container(fuzzer_name)).sas_url
    with tempfile.TemporaryDirectory() as tmpdir:
        command = ["git", "clone", "--depth", "1", fuzzer_url, tmpdir]
        of.logger.info("Dowloading fuzzer '%s' ...", fuzzer_name)
        subprocess.check_call(command)
        if AZCOPY_PATH is None:
            raise Exception("missing azcopy")
        command = [AZCOPY_PATH, "sync", tmpdir, fuzzer_sas]
        of.logger.info(
            "Uploading fuzzer '%s' to OneFuzz: %s",
            fuzzer_name,
            " ".join(command),
        )
        subprocess.check_call(command)


def upload_to_setup_container(of: Onefuzz, helper: JobHelper, setup_dir: str) -> None:
    setup_sas = of.containers.get(helper.containers[ContainerType.setup]).sas_url
    if AZCOPY_PATH is None:
        raise Exception("missing azcopy")
    command = [AZCOPY_PATH, "sync", setup_dir, setup_sas]
    of.logger.info("Uploading '%s' to OneFuzz: %s", setup_dir, " ".join(command))
    subprocess.check_call(command)


def main() -> None:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("setup_dir", type=str, help="Target setup directory")
    parser.add_argument(
        "target_exe", type=str, help="Target executable within setup directory"
    )
    parser.add_argument("build", type=str, help="Target build version.")
    parser.add_argument("pool_name", type=str, help="Worker pool")
    parser.add_argument("--project", type=str, default="msedge", help="Name of project")
    parser.add_argument("--name", type=str, default="browser", help="Name of target")
    parser.add_argument(
        "--duration", type=int, default=1, help="Hours to run the fuzzing task"
    )
    parser.add_argument("--vm_sku", default="Standard_DS1_v2", help="VM image to use")
    parser.add_argument("--notification_config", help="Notification configuration")
    parser.add_argument(
        "--platform",
        type=OS,
        help="Specify Platform. Possible values: %s" % ", ".join([x.name for x in OS]),
    )
    args = parser.parse_args()

    notification_config: Optional[NotificationConfig] = None
    if args.notification_config:
        with open(args.notification_config) as handle:
            notification_config = NotificationConfig.parse_obj(json.load(handle))

    of = Onefuzz()
    logging.basicConfig(level=logging.WARNING)
    of.logger.setLevel(logging.INFO)

    if not os.path.exists(args.target_exe):
        logging.warning(
            "target (%s) does not exist.  Unless this is "
            "downloaded via in-VM setup script, fuzzing will fail",
            args.target_exe,
        )

        if args.platform is None:
            logging.error("Without target exe, platform must be set")
            sys.exit(1)

    helper = JobHelper(
        of,
        of.logger,
        args.project,
        args.name,
        args.build,
        args.duration,
        pool_name=args.pool_name,
        target_exe=args.target_exe,
        platform=args.platform,
    )

    upload_to_fuzzer_container(of, FUZZER_NAME, FUZZER_URL)

    helper.define_containers(
        ContainerType.setup,
        ContainerType.readonly_inputs,
        ContainerType.crashes,
        ContainerType.unique_reports,
        ContainerType.reports,
    )
    helper.create_containers()
    helper.setup_notifications(notification_config)
    upload_to_setup_container(of, helper, args.setup_dir)
    add_setup_script(of, helper.containers[ContainerType.setup])

    containers = [
        (ContainerType.setup, helper.containers[ContainerType.setup]),
        (ContainerType.crashes, helper.containers[ContainerType.crashes]),
        (ContainerType.reports, helper.containers[ContainerType.reports]),
        (ContainerType.unique_reports, helper.containers[ContainerType.unique_reports]),
    ]

    of.logger.info("Creating generic_crash_report task")
    of.tasks.create(
        helper.job.job_id,
        TaskType.generic_crash_report,
        helper.setup_relative_blob_name(args.target_exe, args.setup_dir),
        containers,
        pool_name=args.pool_name,
        duration=args.duration,
    )

    containers = [
        (ContainerType.tools, Container(FUZZER_NAME)),
        (ContainerType.setup, helper.containers[ContainerType.setup]),
        (ContainerType.crashes, helper.containers[ContainerType.crashes]),
        (
            ContainerType.readonly_inputs,
            helper.containers[ContainerType.readonly_inputs],
        ),
    ]

    of.logger.info("Creating generic_generator")

    target_command = [
        "--no-sandbox",
        "--no-first-run",
        "--no-default-browser-check",
        "--allow-file-access-from-files",
        "--disable-popup-blocking",
        "--enable-logging=stderr",
        "--js-flags='--expose_gc'",
        "--window-size=1024,1024",
        "--enable-webgl-draft-extensions",
        "--enable-experimental-web-platform-features",
        "--enable-experimental-canvas-features",
        "--user-data-dir=c:\\msedge-data",
        "{input}",
    ]
    fuzzer_generator = "python.exe"
    fuzzer_generator_options = [
        "{tools_dir}/generator.py",
        "--output_dir",
        "{generated_inputs}",
        "--no_of_files",
        "1000",
    ]

    of.tasks.create(
        helper.job.job_id,
        TaskType.generic_generator,
        helper.setup_relative_blob_name(args.target_exe, args.setup_dir),
        containers,
        pool_name=args.pool_name,
        target_options=target_command,
        duration=args.duration,
        generator_exe=fuzzer_generator,
        generator_options=fuzzer_generator_options,
        reboot_after_setup=True,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json
from typing import Dict, List, Optional

from onefuzztypes.enums import ContainerType, TaskDebugFlag, TaskType
from onefuzztypes.models import NotificationConfig, RegressionReport
from onefuzztypes.primitives import Container, Directory, File, PoolName

from onefuzz.api import Command

from . import JobHelper


class Regression(Command):
    """Regression job"""

    def _check_regression(self, container: Container, file: File) -> bool:
        content = self.onefuzz.containers.files.get(Container(container), file)
        as_str = content.decode()
        as_obj = json.loads(as_str)
        report = RegressionReport.parse_obj(as_obj)

        if report.crash_test_result.crash_report is not None:
            return True

        if report.crash_test_result.no_repro is not None:
            return False

        raise Exception("invalid crash report")

    def generic(
        self,
        project: str,
        name: str,
        build: str,
        pool_name: PoolName,
        *,
        reports: Optional[List[str]] = None,
        crashes: Optional[List[File]] = None,
        target_exe: File = File("fuzz.exe"),
        tags: Optional[Dict[str, str]] = None,
        notification_config: Optional[NotificationConfig] = None,
        target_env: Optional[Dict[str, str]] = None,
        setup_dir: Optional[Directory] = None,
        reboot_after_setup: bool = False,
        target_options: Optional[List[str]] = None,
        dryrun: bool = False,
        duration: int = 24,
        crash_report_timeout: Optional[int] = None,
        debug: Optional[List[TaskDebugFlag]] = None,
        check_retry_count: Optional[int] = None,
        check_fuzzer_help: bool = True,
        delete_input_container: bool = True,
        check_regressions: bool = False,
    ) -> None:
        """
        generic regression task

        :param File crashes: Specify crashing input files to check in the regression task
        :param str reports: Specify specific report names to verify in the regression task
        :param bool check_regressions: Specify if exceptions should be thrown on finding crash regressions
        :param bool delete_input_container: Specify wether or not to delete the input container
        """

        self._create_job(
            TaskType.generic_regression,
            project,
            name,
            build,
            pool_name,
            crashes=crashes,
            reports=reports,
            target_exe=target_exe,
            tags=tags,
            notification_config=notification_config,
            target_env=target_env,
            setup_dir=setup_dir,
            reboot_after_setup=reboot_after_setup,
            target_options=target_options,
            dryrun=dryrun,
            duration=duration,
            crash_report_timeout=crash_report_timeout,
            debug=debug,
            check_retry_count=check_retry_count,
            check_fuzzer_help=check_fuzzer_help,
            delete_input_container=delete_input_container,
            check_regressions=check_regressions,
        )

    def libfuzzer(
        self,
        project: str,
        name: str,
        build: str,
        pool_name: PoolName,
        *,
        reports: Optional[List[str]] = None,
        crashes: Optional[List[File]] = None,
        target_exe: File = File("fuzz.exe"),
        tags: Optional[Dict[str, str]] = None,
        notification_config: Optional[NotificationConfig] = None,
        target_env: Optional[Dict[str, str]] = None,
        setup_dir: Optional[Directory] = None,
        reboot_after_setup: bool = False,
        target_options: Optional[List[str]] = None,
        dryrun: bool = False,
        duration: int = 24,
        crash_report_timeout: Optional[int] = None,
        debug: Optional[List[TaskDebugFlag]] = None,
        check_retry_count: Optional[int] = None,
        check_fuzzer_help: bool = True,
        delete_input_container: bool = True,
        check_regressions: bool = False,
    ) -> None:

        """
        libfuzzer regression task

        :param File crashes: Specify crashing input files to check in the regression task
        :param str reports: Specify specific report names to verify in the regression task
        :param bool check_regressions: Specify if exceptions should be thrown on finding crash regressions
        :param bool delete_input_container: Specify wether or not to delete the input container
        """

        self._create_job(
            TaskType.libfuzzer_regression,
            project,
            name,
            build,
            pool_name,
            crashes=crashes,
            reports=reports,
            target_exe=target_exe,
            tags=tags,
            notification_config=notification_config,
            target_env=target_env,
            setup_dir=setup_dir,
            reboot_after_setup=reboot_after_setup,
            target_options=target_options,
            dryrun=dryrun,
            duration=duration,
            crash_report_timeout=crash_report_timeout,
            debug=debug,
            check_retry_count=check_retry_count,
            check_fuzzer_help=check_fuzzer_help,
            delete_input_container=delete_input_container,
            check_regressions=check_regressions,
        )

    def _create_job(
        self,
        task_type: TaskType,
        project: str,
        name: str,
        build: str,
        pool_name: PoolName,
        *,
        crashes: Optional[List[File]] = None,
        reports: Optional[List[str]] = None,
        target_exe: File = File("fuzz.exe"),
        tags: Optional[Dict[str, str]] = None,
        notification_config: Optional[NotificationConfig] = None,
        target_env: Optional[Dict[str, str]] = None,
        setup_dir: Optional[Directory] = None,
        reboot_after_setup: bool = False,
        target_options: Optional[List[str]] = None,
        dryrun: bool = False,
        duration: int = 24,
        crash_report_timeout: Optional[int] = None,
        debug: Optional[List[TaskDebugFlag]] = None,
        check_retry_count: Optional[int] = None,
        check_fuzzer_help: bool = True,
        delete_input_container: bool = True,
        check_regressions: bool = False,
    ) -> None:

        if dryrun:
            return None

        self.logger.info("creating regression task from template")

        helper = JobHelper(
            self.onefuzz,
            self.logger,
            project,
            name,
            build,
            duration,
            pool_name=pool_name,
            target_exe=target_exe,
        )

        helper.define_containers(
            ContainerType.setup,
            ContainerType.crashes,
            ContainerType.reports,
            ContainerType.no_repro,
            ContainerType.unique_reports,
            ContainerType.regression_reports,
        )

        containers = [
            (ContainerType.setup, helper.containers[ContainerType.setup]),
            (ContainerType.crashes, helper.containers[ContainerType.crashes]),
            (ContainerType.reports, helper.containers[ContainerType.reports]),
            (ContainerType.no_repro, helper.containers[ContainerType.no_repro]),
            (
                ContainerType.unique_reports,
                helper.containers[ContainerType.unique_reports],
            ),
            (
                ContainerType.regression_reports,
                helper.containers[ContainerType.regression_reports],
            ),
        ]

        if crashes:
            helper.containers[
                ContainerType.readonly_inputs
            ] = helper.get_unique_container_name(ContainerType.readonly_inputs)
            containers.append(
                (
                    ContainerType.readonly_inputs,
                    helper.containers[ContainerType.readonly_inputs],
                )
            )

        helper.create_containers()
        if crashes:
            for file in crashes:
                self.onefuzz.containers.files.upload_file(
                    helper.containers[ContainerType.readonly_inputs], file
                )

        helper.setup_notifications(notification_config)

        helper.upload_setup(setup_dir, target_exe)
        target_exe_blob_name = helper.setup_relative_blob_name(target_exe, setup_dir)

        self.logger.info("creating regression task")
        task = self.onefuzz.tasks.create(
            helper.job.job_id,
            task_type,
            target_exe_blob_name,
            containers,
            pool_name=pool_name,
            duration=duration,
            vm_count=1,
            reboot_after_setup=reboot_after_setup,
            target_options=target_options,
            target_env=target_env,
            tags=tags,
            target_timeout=crash_report_timeout,
            check_retry_count=check_retry_count,
            debug=debug,
            check_fuzzer_help=check_fuzzer_help,
            report_list=reports,
        )
        helper.wait_for_stopped = check_regressions

        self.logger.info("done creating tasks")
        helper.wait()

        if check_regressions:
            task = self.onefuzz.tasks.get(task.task_id)
            if task.error:
                raise Exception("task failed: %s", task.error)

            container = helper.containers[ContainerType.regression_reports]
            for filename in self.onefuzz.containers.files.list(container).files:
                self.logger.info("checking file: %s", filename)
                if self._check_regression(container, File(filename)):
                    raise Exception(f"regression identified: {filename}")
            self.logger.info("no regressions")

        if (
            delete_input_container
            and ContainerType.readonly_inputs in helper.containers
        ):
            helper.delete_container(helper.containers[ContainerType.readonly_inputs])



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os

from PyInstaller.utils.hooks import get_module_file_attribute

res_loc = os.path.dirname(get_module_file_attribute("wcwidth"))
datas = [
    (os.path.join(res_loc, "version.json"), "wcwidth"),
]



import logging
import os

import azure.functions as func

from onefuzz.api import Onefuzz


def main(req: func.HttpRequest) -> func.HttpResponse:
    logging.info("Python HTTP trigger function processed a request.")

    o = Onefuzz()
    o.config(
        endpoint=os.environ.get("ONEFUZZ_ENDPOINT"),
        authority=os.environ.get("ONEFUZZ_AUTHORITY"),
        client_id=os.environ.get("ONEFUZZ_CLIENT_ID"),
        client_secret=os.environ.get("ONEFUZZ_CLIENT_SECRET"),
    )
    info = o.info.get()
    return func.HttpResponse(info.json())



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import json
import logging
import os
import tempfile
from typing import Optional

from onefuzztypes.enums import ContainerType, TaskType
from onefuzztypes.models import NotificationConfig
from onefuzztypes.primitives import Container

from onefuzz.api import Onefuzz
from onefuzz.templates import JobHelper

SETUP = """
cd /onefuzz
apt-get update
apt-get -y install build-essential libbfd-dev libunwind-dev
git clone https://github.com/google/honggfuzz/
cd /onefuzz/honggfuzz
make
"""


def add_setup_script(of: Onefuzz, container: Container) -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        setup = os.path.join(tmpdir, "setup.sh")
        with open(setup, "w") as handle:
            handle.write(SETUP)
        of.containers.files.upload_file(container, setup)


def main() -> None:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("setup_dir", type=str, help="Target setup directory")
    parser.add_argument(
        "target_exe", type=str, help="Target executable within setup directory"
    )
    parser.add_argument("project", type=str, help="Name of project")
    parser.add_argument("name", type=str, help="Name of target")
    parser.add_argument("build", type=str, help="Target build version.")
    parser.add_argument("pool_name", type=str, help="VM pool to use")
    parser.add_argument(
        "--duration", type=int, default=1, help="Hours to run the fuzzing task"
    )
    parser.add_argument("--inputs", help="seeds to use")
    parser.add_argument("--notification_config", help="Notification configuration")
    args = parser.parse_args()

    notification_config: Optional[NotificationConfig] = None
    if args.notification_config:
        with open(args.notification_config) as handle:
            notification_config = NotificationConfig.parse_obj(json.load(handle))

    of = Onefuzz()
    logging.basicConfig(level=logging.WARNING)
    of.logger.setLevel(logging.DEBUG)

    helper = JobHelper(
        of,
        of.logger,
        args.project,
        args.name,
        args.build,
        args.duration,
        pool_name=args.pool_name,
        target_exe=args.target_exe,
    )

    helper.define_containers(
        ContainerType.setup,
        ContainerType.readonly_inputs,
        ContainerType.crashes,
        ContainerType.inputs,
        ContainerType.reports,
        ContainerType.unique_reports,
    )
    helper.create_containers()
    helper.setup_notifications(notification_config)
    helper.upload_setup(args.setup_dir, args.target_exe)
    if args.inputs:
        helper.upload_inputs(args.inputs)

    add_setup_script(of, helper.containers[ContainerType.setup])

    containers = [
        (ContainerType.setup, helper.containers[ContainerType.setup]),
        (ContainerType.crashes, helper.containers[ContainerType.crashes]),
        (ContainerType.reports, helper.containers[ContainerType.reports]),
        (ContainerType.unique_reports, helper.containers[ContainerType.unique_reports]),
    ]

    of.logger.info("Creating generic_crash_report task")
    of.tasks.create(
        helper.job.job_id,
        TaskType.generic_crash_report,
        helper.setup_relative_blob_name(args.target_exe, args.setup_dir),
        containers,
        pool_name=args.pool_name,
        duration=args.duration,
    )

    containers = [
        (ContainerType.tools, Container("honggfuzz")),
        (ContainerType.setup, helper.containers[ContainerType.setup]),
        (ContainerType.crashes, helper.containers[ContainerType.crashes]),
        (
            ContainerType.inputs,
            helper.containers[ContainerType.inputs],
        ),
    ]

    supervisor_options = [
        "-n1",
        "--crashdir",
        "{crashes}",
        "-u",
        "-i",
        "{input_corpus}",
        "--",
        "{target_exe}",
        "{input}",
    ]

    of.tasks.create(
        helper.job.job_id,
        TaskType.generic_supervisor,
        helper.setup_relative_blob_name(args.target_exe, args.setup_dir),
        containers,
        pool_name=args.pool_name,
        supervisor_exe="/onefuzz/honggfuzz/honggfuzz",
        supervisor_options=supervisor_options,
        supervisor_input_marker="___FILE___",
        duration=args.duration,
        vm_count=1,
        tags=helper.tags,
    )


if __name__ == "__main__":
    main()
